/* 
compile with -DTTY to have it output text-only tree without GUI

compile with -DADDROOT to have it root the unrooted input tree and write tree on stdout (no GUI)

compile with -DNO_GUI to have it work postscript_only without GUI at all

WIN_MOTIF resources:
Vibrant.systemfont : helvetica,14,b
Njplot.systemfont : times,14,b
Vibrant.programfont : fixed,14
placed in file
$HOME/.Xdefaults
$XAPPLRESDIR/Vibrant
/usr/X11R6/lib/X11/app-defaults/Vibrant    (Linux)
/usr/openwin/lib/X11/app-defaults/Vibrant   (Solaris)
*/
#include <math.h>
#include <time.h>
#include <ctype.h>

#define MAX_FRAC 0.95

#if defined(TTY) || defined (ADDROOT)
#define NO_GUI
#endif
#ifdef NO_GUI

#define FALSE 0
#define TRUE (!FALSE)
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#else

#include <vibtypes.h>
#include <vibprocs.h>

#if defined(WIN_MSWIN) || defined (WIN_MAC)
#define  MICRO
#endif

#ifdef WIN_MAC

#ifdef verify
#undef verify
#endif
#include <Carbon.h>
//#include <MacHeadersCarbon.h>  // precompiled version
#define myWaitAndProcessNextEvent Nlm_ProcessAnEvent
extern int Nlm_textScrapFull;

#elif defined(WIN_MSWIN)
#define HDC void *
extern HDC Nlm_currentHDC;
extern HDC Nlm_GetPicWinHDC ( void );
char *get_prog_dir(void);
#define myWaitAndProcessNextEvent Nlm_ProcessAnEvent

#else  /* unix */

#undef Boolean
#include <X11/Intrinsic.h>

extern XtAppContext Nlm_appContext;
extern Nlm_CharPtr Nlm_XrmGetResource(const Nlm_Char PNTR _resource);
void set_systemfont(char *systemfont);

void myWaitAndProcessNextEvent (void)
{
XEvent       event;
if( XtAppPending(Nlm_appContext) & XtIMXEvent ) {
	Nlm_ProcessAnEvent();
	}
else	{
/* wait until next event and catch it */
    	XtAppNextEvent (Nlm_appContext, &event);
    	XtDispatchEvent (&event); 
	}
}

#endif


#include <document.h> /* need this include after the other mac includes */

#ifdef WIN_MAC /* need the mac version of each of these functions */
#undef MoveTo
#undef LineTo
#undef ClipRect
#undef TextWidth
#undef DrawString
#endif

struct win_help_extra {
	DoC document;
	FonT police;
	};

struct my_RecT4 {
	Nlm_Int4 left, top, right, bottom;
	}  print_rect;

typedef struct {
	Nlm_WindoW nj_plot;
	Nlm_IteM open_button;
	Nlm_PrompT click_message;
	Nlm_MenU tree_font_menu;
	Nlm_MenU paper_menu;
	Nlm_IteM save_plot_button;
	Nlm_IteM save_tree_button;
	Nlm_IteM SearchItem;
	Nlm_IteM AgainItem;
	Nlm_ChoicE choix_font, choix_taille;
	Nlm_IteM bold_item, italic_item;
	Nlm_ButtoN branch_length_button;
	Nlm_ButtoN bootstrap_button;
	Nlm_GrouP choix_group;
	Nlm_ButtoN reroot_button;
	Nlm_ButtoN subtree_up_button;
	Nlm_IteM exit_button;
	Nlm_PaneL tree_plot;
	float zoomvalue;
	Nlm_PrompT zoomprompt;
	char *tree_name;
} FD_nj_plot;

#endif /* NO_GUI */


/* typedefs */

typedef enum { times = 1, helvetica, courier} font_name;
typedef enum { font_tiny = 1, font_small,
	font_normal, font_medium, font_large, font_bold, font_italic } font_size;
typedef enum { A4 = 1, LETTER } 
	paper_item;

struct noeud {
	double l1,l2,l3;
	struct noeud *v1,*v2,*v3;
	} **tabtax;
	
typedef struct { /* une branche definie par ses deux extremites */
	struct noeud *bouta;
	struct noeud *boutb;
	char *br_label;
	} branche;
branche *branches;

#define s_noeud sizeof(struct noeud)
char **tabnames, **labels, *br_length_txt, *end_br_length;
struct nom {
	double x,y;
	char *nom;
	char disp_option;
	} *noms;
struct mon_point {
	double x,y;
	int number;
	} *points;
struct trait {
	double xd,yd,xf,yf;
	} *traits;
struct noeud *racine;

typedef struct _moments {
	int N; /* nbre de fils */
	double somme, carres; /* somme des longueurs jusqu'aux feuilles, et somme des carres */
	} moments;



/* prototypes of included functions */
#ifndef NO_GUI
FD_nj_plot *create_win_nj_plot(void);
void cre_help_win(IteM item);
FILE *open_path(char *fname);
void topic_callback(LisT list);
void help_ok_callback(ButtoN bouton);
#ifdef WIN_MAC
void load_help_topics(char *h, LisT topic_list);
#else
void load_help_topics(FILE *fich, LisT topic_list);
#endif
void subtree_up_callback(ButtoN bouton);
 void open_callback(Nlm_IteM );
 void font_callback(Nlm_ChoicE);
 void paper_callback(Nlm_ChoicE);
 void save_plot_callback(Nlm_IteM);
 void save_tree_callback(Nlm_IteM);
 void toggle_branch_callback(Nlm_ButtoN);
 void toggle_bootstrap_callback(Nlm_ButtoN);
 void operation_callback(Nlm_GrouP);
 void subtree_callback(void);
 void swap_callback(void);
 void new_outgroup_callback(void);
 void show_tree_callback(void);
 void exit_callback(Nlm_IteM);
 void tree_draw_proc(Nlm_PaneL);
 void tree_click_proc(Nlm_PaneL, Nlm_PoinT);
 int myrint(double val);
static void quitproc(WindoW);
void win_resize_proc(WindoW);
void change_panel_size(PaneL panel, int change_x, int change_y);
void string_callback(ButtoN item);
void search_callback(IteM item);
void process_keys(Nlm_Char key);
void change_page_count(ButtoN item);
void page_count_callback(IteM item);
void clear_tree(IteM);
void scrollnotu(int count);
void paste_tree(IteM);
void scrollcallback(Nlm_BaR b, Nlm_SlatE g, Nlm_Int2 before, Nlm_Int2 after);
void zoomcallback(Nlm_SwitcH s, Nlm_Int2 before, Nlm_Int2 after);
#ifdef MICRO
void copy_plot(IteM);
void print_title(int x, int y, char *text, Nlm_FonT title_font, int p,int totp);
void print_plot(IteM);
#endif
#ifdef WIN_MAC
extern void add_apropos(char *);
int is_macosx(void);
int crefpict(char *, PicHandle );
void example_tree(IteM unused);
void pict_plot(void);
extern int MG_GetInputFName(char *fname, int maxl);
extern int MG_GetOutputFName(char *fname, int maxl, char *dfault);
#else 
#define MG_GetOutputFName Nlm_GetOutputFileName
#define MG_GetInputFName(a,b) Nlm_GetInputFileName(a, b, NULL, NULL)
#endif

#endif /* NO_GUI */

void process_args(int *argc, char *argv[], char **pred);
void prepare_fonts(void);
void prepare_ps_font(int fontnum, int use_bold, int use_italic);
int direct_ps_plot(void);
int calc_otu(struct noeud *pere, struct noeud *centre);
void do_plot(void);
int calc_text_size(char *text, int *pheight, int *pascent);
void plotstring(char *nom);
void dir_moveto(double x,double y);
void dir_lineto(double x,double y);
void scale_window(double lxmin, double lxmax, double lymin, double lymax,
 	double pxmin, double pxmax, double pymin, double pymax);
void ch_echelle(double lx, double ly, double *px, double *py);
void convert_mem_point(void);
void init_tree(char *titre);
char *preptree(char *fname);
char *check_alloc(int nbrelt, int sizelt);
void err_message(const char *text);
void loadphylip(char *arbre);
struct noeud *unrootedset(char *deb, char *fin, branche **p_int_br);
char *extract_filename(char *name);
char *nextpar(char *pospar);
void make_binary_or_unrooted(char *arbre);
int make_binary(char *arbre, char *debut, char *fin, int go_down);
void mydrawstring(double x, double y, char *nom, char option);
void moveto(double x,double y);
void lineto(double x,double y);
double tree_bal(struct noeud *centre, struct noeud *p1, struct noeud *p2);
void get_next_br(struct noeud *pere, struct noeud *fils, 
	double **abdown, double **abup, struct noeud **f1, struct noeud **f2);
void place_midpoint_root(void);
void parcourir_branches(struct noeud *centre, struct noeud *origine);
void process_branche(struct noeud *cote1, struct noeud *cote2, double length);
double get_length_down(struct noeud *, struct noeud *);
moments stat_from_node(struct noeud *pere, struct noeud *racine);
void runtree(void);
void mem_plot(struct noeud *pere, struct noeud *centre, double currx,
	double *curry);
void mem_point(double x, double y, int number);
void mem_nom(double x, double y, char *nom, char option);
void mem_trait(double xd, double yd, double xf, double yf);
char *get_br_label(struct noeud *a, struct noeud *b);
int get_br_from_bouts(struct noeud *a, struct noeud *b);
void free_tree(void);
char *ecrit_arbre_parenth(struct noeud *root);
char *recur_ecrit_arbre(struct noeud *centre, char *arbre, char *finarbre);
void removeroot(void);
void remove_arg(int target, int *argc, char *argv[]);
double length_log_phys(double p);
double length_phys_log(double p);
double arrondi_echelle(double x);
double calc_echelle(double larg);
void draw_scale(void);
int calc_n_desc(struct noeud *pere);
void bad_format(char *);
void majuscules(char *p);
void postscript_plot(void);
void tty_plot(void);




/* globals */
#ifndef NO_GUI
FD_nj_plot *fd_nj_plot;
Nlm_FonT current_font;
#endif
int page_count = 1;
double	physx,physy,physx_min,physy_min,physx_corr,physy_corr;
/* window scaling variables */
double tek_xmin,tek_xmax,tek_ymin,tek_ymax,tek_dx,tek_dy;
double tek_minx,tek_maxx,tek_miny,tek_maxy;
int char_height, ascent;
int *widnames;
double *profs;
double maxx, maxy, deltay, root_br_l, nexty;
/* max length of taxa names */
int tottraits, ps_plot, has_br_length = 0, totnoms = -1, notu, totbranches, 
	num_noeud, rooted, nextotu, totpoints, no_title = FALSE,
	root_num, swap = 0, plot_br_l = 0, show_bootstrap = FALSE, has_internal = FALSE;
char current_ps_font[40];
int current_font_size_rank = font_normal;
#ifdef WIN_MSWIN
int list_font_size[] = { 0, 6, 8, 10, 12, 14, 18 };
char list_font_name[4][40] = {"", "Times New Roman", "Lucida Sans", 
	"Courier New"};
#else
int list_font_size[] = { 0, 8, 10, 12, 14, 18 };
char list_font_name[4][15] = {"", "Times", "Helvetica", "Courier"};
#endif
char list_ps_font_name[4][15] = {"", "Times", "Helvetica", "Courier"};
paper_item paper_choice = A4;
char fname[200], plotfilename[200], rooted_fname[200];
FILE *plotfile;
#ifdef WIN_MAC
char *help_file, *example;
#else
FILE *help_file;
#endif
int long_arbre_parenth; /* long de la chaine decrivant l'arbre lu */
int need_runtree;
int converted;
int ps_width = 500, ps_height = 0 /* 0 means not defined by this variable */;
enum operations {show_tree = 1, depl_racine, permutation, subtree} choix;
/* variables globales pour memoriser la meilleure branche:
celle qui partage l'arbre en 2 parties les plus egales possibles en prof moyenne
*/
double current_best_diff, current_br_length, current_balance;
struct noeud *current_cote1=NULL, *current_cote2=NULL;
#define VERY_BIG 9.e99

struct noeud *subtree_center, *subtree_ascend;
int subtree_notu;
int doing_copy = FALSE;
int doing_print = FALSE;
float postscript_ratio = 0.60;

char **tty_page;
int tty_x, tty_y;
char *outotu = NULL; /* value of optional argument -outotu name */


#ifdef WIN_MAC
Rect		myrect;
PicHandle 	mypicture;
int 		margin;
#endif

#ifdef NO_GUI
int ps_plot_only = TRUE;
int main(int argc, char **argv)
{
char *red;
#else
int ps_plot_only = FALSE;

#ifdef hpux
extern int argc;
extern char **argv;
#else
int argc;
char **argv;
#endif

Nlm_Int2 Nlm_Main(void)
{
char *red;
#ifndef hpux
argc = Nlm_GetArgc();
argv = Nlm_GetArgv();
#endif
#endif /* NO_GUI */

*fname = 0;
process_args(&argc, argv, &red);
if(argc>=2) {
	strcpy(fname,argv[1]);
	}
else	{
	*fname=0;
	}
#ifdef unix
putenv("LC_NUMERIC=C");
#endif
   if(ps_plot_only) {
 	int status = 0;
 	if( (int) strlen(fname) > 0) status = direct_ps_plot();
	return status;
	}
#ifndef NO_GUI
   Nlm_WatchCursor();
#ifdef WIN_MOTIF
	{char *font;
	font = Nlm_XrmGetResource("systemfont");
	if(font != NULL) set_systemfont(font);
	font = Nlm_XrmGetResource("programfont");
	if(font != NULL) Nlm_programFont = Nlm_ParseFont(font); 
	}
#endif
   fd_nj_plot = create_win_nj_plot();
   prepare_fonts();
   fd_nj_plot->tree_name = fname;
   if(*fname != 0) init_tree( extract_filename(fname) );
   else SetTitle(fd_nj_plot->nj_plot, "njplot");
   Show(fd_nj_plot->nj_plot);
   {
	RecT *p_rect;
   	p_rect = (RecT *)GetWindowExtra(fd_nj_plot->nj_plot);
   	ObjectRect(fd_nj_plot->nj_plot, p_rect);
   }
   if(red != NULL) search_callback( (Nlm_IteM)red );
   Nlm_ArrowCursor();
   ProcessEvents();
#endif
   return 0;
}


#ifndef NO_GUI
FD_nj_plot *create_win_nj_plot(void)
{
  WindoW win;
  MenU menu;
  ChoicE choix_paper;
  GrouP group;
  FD_nj_plot *fdui = (FD_nj_plot *) calloc(1, sizeof(FD_nj_plot));
  static RecT winrect;
  PoinT position;
Nlm_Handle obj; int left, right;
  
#ifdef WIN_MAC
long gval;
add_apropos("njplot");
#endif
Nlm_KeyboardView(process_keys);
fdui->nj_plot = win = DocumentWindow(-50, -33, -3, -3, "Njplot", 
		quitproc, win_resize_proc);
#ifdef WIN_MAC
#define OPEN_LABEL      "Open        O"
#define PRINT_LABEL     "Print         P"
#define SAVE_PLOT_LABEL "Save Plot   S"
#define QUIT_LABEL      "Quit          Q"
#define COPY_LABEL      "Copy   C"
#define PASTE_LABEL     "Paste  V"
#define FIND_LABEL      "Find    F"
#define AGAIN_LABEL     "Again  A"
#define PLACE_OF_MENUS NULL
#else
#define OPEN_LABEL      "Open        ^O"
#define PRINT_LABEL     "Print       ^P"
#define SAVE_PLOT_LABEL "Save plot  ^S"
#define QUIT_LABEL      "Quit          ^Q"
#define COPY_LABEL      "Copy     ^C"
#define PASTE_LABEL     "Paste     ^V"
#define FIND_LABEL      "Find       ^F"
#define AGAIN_LABEL     "Again     ^A"
#define PLACE_OF_MENUS fdui->nj_plot
#endif
  menu = PulldownMenu(PLACE_OF_MENUS, "File");
  fdui->open_button =      CommandItem(menu, OPEN_LABEL, open_callback);
  fdui->save_plot_button = CommandItem(menu, SAVE_PLOT_LABEL, save_plot_callback);
  fdui->save_tree_button = CommandItem(menu, "Save Tree", save_tree_callback);
#ifdef MICRO
  CommandItem(menu, PRINT_LABEL, print_plot);
#endif
#ifdef WIN_MAC
if( ! is_macosx() )
#endif
  CommandItem(menu, QUIT_LABEL, exit_callback);
  Advance(PLACE_OF_MENUS);
  menu = PulldownMenu(PLACE_OF_MENUS, "Edit");
  CommandItem(menu, "Clear", clear_tree);
#ifdef MICRO
  CommandItem(menu, COPY_LABEL, copy_plot);
#endif
  CommandItem(menu, PASTE_LABEL, paste_tree);
#ifdef WIN_MAC
  SeparatorItem(menu);
  CommandItem(menu, "Example", example_tree);
#endif
  SeparatorItem(menu);
  fdui->SearchItem = CommandItem(menu, FIND_LABEL, search_callback);
  fdui->AgainItem = CommandItem(menu, AGAIN_LABEL, search_callback);
/*  Disable(fdui->AgainItem); */
  Advance(PLACE_OF_MENUS);
  fdui->tree_font_menu = PulldownMenu(PLACE_OF_MENUS, "Font"); 
  fdui->choix_font = ChoiceGroup(fdui->tree_font_menu, font_callback);
  Nlm_ChoiceItem(fdui->choix_font, list_font_name[1]);
  Nlm_ChoiceItem(fdui->choix_font, list_font_name[2]);
  Nlm_ChoiceItem(fdui->choix_font, list_font_name[3]);
  SetValue(fdui->choix_font, times);
  SeparatorItem(fdui->tree_font_menu);
  fdui->choix_taille = ChoiceGroup(fdui->tree_font_menu, font_callback);
#ifdef WIN_MSWIN
  ChoiceItem(fdui->choix_taille, "6");
#endif
  ChoiceItem(fdui->choix_taille, "8");
  ChoiceItem(fdui->choix_taille, "10");
  ChoiceItem(fdui->choix_taille, "12");
  ChoiceItem(fdui->choix_taille, "14");
  ChoiceItem(fdui->choix_taille, "18");
  SetValue(fdui->choix_taille, current_font_size_rank);
  SeparatorItem(fdui->tree_font_menu);
  fdui->bold_item = StatusItem(fdui->tree_font_menu, "Bold", 
  	(Nlm_ItmActnProc) font_callback);
  fdui->italic_item = StatusItem(fdui->tree_font_menu, "Italic", 
  	(Nlm_ItmActnProc) font_callback);
  SetStatus(fdui->bold_item, FALSE);
  SetStatus(fdui->italic_item, FALSE);
  
  Advance(PLACE_OF_MENUS);
  fdui->paper_menu = PulldownMenu(PLACE_OF_MENUS, "Paper"); 
  choix_paper = ChoiceGroup(fdui->paper_menu, paper_callback);
  ChoiceItem(choix_paper, "A4");
  ChoiceItem(choix_paper, "US letter");
  SetValue(choix_paper, paper_choice);
  SeparatorItem(fdui->paper_menu);
#ifdef WIN_MSWIN /* SetTitle ne marche pas sur PC! */
  	CommandItem(fdui->paper_menu, "Page count", page_count_callback);
#elif defined(WIN_MAC) /* en 2 temps car () ds menu ne marche pas sur Mac ! */
  {	Nlm_IteM item; char nom[25];
  	item = CommandItem(fdui->paper_menu, "tmp", page_count_callback);
	sprintf(nom, "Page count (%d)", page_count);
  	SetTitle(item, nom);
  }
#else /* unix */
  {	char nom[25];
	sprintf(nom, "Page count (%d)", page_count);
  	CommandItem(fdui->paper_menu, nom, page_count_callback);
  }
#endif  

  Advance(PLACE_OF_MENUS);
  obj = PulldownMenu(PLACE_OF_MENUS, "Help");
  CommandItem(obj, "Help", cre_help_win);

  Break(PLACE_OF_MENUS);
  fdui->choix_group = NormalGroup(win, 4, 1, "Operation", programFont, operation_callback);
  RadioButton(fdui->choix_group, "Full tree");
  fdui->reroot_button = RadioButton(fdui->choix_group, "New outgroup");
  RadioButton(fdui->choix_group, "Swap nodes");
  RadioButton(fdui->choix_group, "Subtree");
  SetValue(fdui->choix_group, show_tree);

  Break(win);
  group = NormalGroup(win, 2, 1, "Display", programFont, NULL);
  fdui->branch_length_button = CheckBox(group, "Branch lengths", 
		toggle_branch_callback);
  SetStatus(fdui->branch_length_button, plot_br_l);
  fdui->bootstrap_button = CheckBox(group, "Bootstrap values", 
			toggle_bootstrap_callback);
  SetStatus(fdui->bootstrap_button, show_bootstrap);

  Advance(win);
  group = NormalGroup(win, 2, 1, "Zoom", programFont, NULL);
obj = Nlm_UpDownSwitch(group, 0, zoomcallback);
fdui->zoomvalue = 1.;
Nlm_SetSwitchParams( (Nlm_SwitcH)obj, 1, 100);
fdui->zoomprompt = Nlm_StaticPrompt(group, "100%", 35, Nlm_stdLineHeight, Nlm_programFont, 'l');

  Advance(win);
  GetNextPosition(win, &position);
  position.x += 5;
  position.y += 10;
  SetNextPosition(win, position);
  obj = fdui->subtree_up_button = PushButton(win, "Subtree Up", subtree_up_callback);
  Disable(fdui->subtree_up_button);
  

Break(win);
Nlm_GetPosition(obj, &winrect); 
right = winrect.right;
Nlm_GetNextPosition(win, &position); 
left = position.x;

#ifdef WIN_MSWIN
#define PLOT_HEIGHT 430
#elif defined(WIN_MAC)
#define PLOT_HEIGHT 300
#else
#define PLOT_HEIGHT 350
#endif

fdui->tree_plot = Nlm_AutonomousPanel(win, right - left + 1, PLOT_HEIGHT, 
		tree_draw_proc, scrollcallback, NULL, sizeof(int), NULL, NULL); 
	{int hidebar = TRUE; /* put TRUE iff scrollbar should be hidden */
	Nlm_SetPanelExtra(fdui->tree_plot, &hidebar);
	}
SetPanelClick(fdui->tree_plot, tree_click_proc, NULL, NULL, NULL);
obj = Nlm_GetSlateVScrollBar((SlatE)fdui->tree_plot);
Nlm_Hide(obj);
Nlm_SetBarValue(obj, 0);
Nlm_SetBarMax(obj, 1);

{int space = 0; /* some space at the bottom of the window */
#if defined(__APPLE__) || defined(WIN_MAC)
space = Nlm_hScrollBarHeight;
#elif defined(WIN_MOTIF)
space = 3;
#endif
if(space > 0) {
	Nlm_Break(win);
	Nlm_StaticPrompt(win, "", 4, space, Nlm_programFont, 'l');
	}
}

winrect.left = -1;
SetWindowExtra(win, &winrect, NULL);
#ifdef WIN_MAC
	{char **h; int l;
	h = GetResource('TEXT', 128);
	if(h == NULL) help_file = NULL;
	else	{
		l = GetHandleSize(h);
		help_file = (char *)malloc(l + 1);
		}
	if(help_file != NULL) {
		HLock(h);
		memcpy(help_file, *h, l);
		HUnlock(h);
		help_file[l] = 0;
		}
	ReleaseResource(h);
	h = GetResource('TEXT', 129);
	if(h == NULL) example = NULL;
	else	{
		l = GetHandleSize(h);
		example = (char *)malloc(l + 1);
		}
	if(example != NULL) {
		HLock(h);
		memcpy(example, *h, l);
		HUnlock(h);
		example[l] = 0;
		}
	ReleaseResource(h);
	}
#else
	help_file = fopen("/usr/share/njplot/njplot.help","r");
#endif
#ifdef WIN_MSWIN
	{ /* contre bug fenetre mange un peu du panel */
	Int2 right;
  	Advance(win);
  	GetNextPosition(win, &position);
  	right = position.x + 5;
  	SetNextPosition(win, position);
  	Break(win);
  	GetNextPosition(win, &position);
  	position.x = right;
  	position.y += 3;
  	SetNextPosition(win, position);
  	StaticPrompt(win, "", 1, 1, programFont, 'l');
  	}
#endif
  return fdui;
}


FILE *open_path(char *fname)  /* to open in read-only file fname searching for 
				 it through all path directories */
{
#define Mxdir 200
        static char dir[Mxdir+1];
        char *path, *deb, *fin;
        FILE *fich;
        int lf, ltot;
#ifdef __VMS
	static char vmspath[] = "seaview";
	path = vmspath;
#elif defined(unix)
        path = getenv("PATH"); 	/* get the list of path directories, 
					separated by :
    				*/
#elif defined(WIN_MSWIN)
/* try first dir where program was launched */
	path = get_prog_dir();
#else
	path = NULL;
#endif
        if (path == NULL ) {
        	fich = fopen(fname,"r");
        	return fich;
        	}
        lf=strlen(fname);
        deb=path;
        do
                {
#ifdef WIN_MSWIN
		fin = NULL;
#else
                fin = strchr(deb,':');
#endif
                if(fin!=NULL)
                        { strncpy(dir,deb,fin-deb); ltot=fin-deb; }
                else
                        { strcpy(dir,deb); ltot=strlen(dir); }
                /* now one directory is in string dir */
                if( ltot + lf + 1 <= Mxdir)
                        {
#ifdef __VMS
                        dir[ltot]=':';
#elif defined(WIN_MSWIN)
			if(dir[ltot-1] == '\\') ltot--;
                        else 	dir[ltot] = '\\';
#else 
                        dir[ltot]='/';
#endif
                        strcpy(dir+ltot+1,fname); /* now dir is appended with filename */
                        fich = fopen(dir,"r");
                        if( fich  != NULL) break;
                        }
                else fich = NULL;
                deb=fin+1;
                }
        while (fin != NULL);
	if(fich == NULL) {
			fich=fopen(fname,"r"); /* try also current directory */
			}
        return fich;
#undef Mxdir
}


void cre_help_win(IteM bouton)
{
static WindoW win;
LisT topic_list;
static DoC help_doc;
static struct win_help_extra win_data;
static int first = TRUE;
int larg;
GrouP group;

if(first) {
	if(help_file == NULL) {
		Disable(bouton);
		err_message(
#ifdef WIN_MAC
	"Sorry, no help resource available."
#else
	"Sorry, help file njplot.help is not found."
#endif
		);
		return;
		}
	first = FALSE;
	win = DocumentWindow( -50, -33, -3, -3, "Njplot Help", (Nlm_WndActnProc)help_ok_callback, NULL);
	group = HiddenGroup(win, -1, -2, NULL);
	SetGroupSpacing(group, 0, 5);
	topic_list = SingleList(group, 7, 8, topic_callback);
	load_help_topics(help_file, topic_list);
	DefaultButton(group, "Ok", help_ok_callback);
	Advance(win);
#ifdef WIN_MSWIN
	win_data.police = ParseFont("Courier New,10");
#else
	win_data.police = programFont;
#endif
	SelectFont(win_data.police);
	larg = 81 * Nlm_TextWidth("Q", 1);
	win_data.document = DocumentPanel(win, larg, 20 * Nlm_LineHeight() );
	SetWindowExtra(win, &win_data, NULL);
	}
Show(win);
Select(win);
}


void topic_callback(LisT list)
{
int num_topic;
char ligne[100], *p, *q;
static int last_topic = 0;
Int2 total;
static char txt[5000];
WindoW win;
DoC help_doc;
struct win_help_extra *p_win_data;

num_topic = GetValue(list);
if(num_topic == 0 || num_topic == last_topic) return;
last_topic = num_topic;
win = ParentWindow(list);
p_win_data = (struct win_help_extra *)GetWindowExtra(win);
help_doc = p_win_data->document;
Hide(help_doc);
GetDocParams(help_doc, &total, NULL);
while(total > 0) DeleteItem(help_doc, total--);
#ifdef WIN_MAC
p = help_file;
while(num_topic > 0) {
	if(strncmp(p, ">>>", 3) == 0) --num_topic;
	p = strchr(p, '\r') + 1;
	}
#else
rewind(help_file);
while(fgets(ligne, sizeof(ligne), help_file) != NULL) {
	if(strncmp(ligne, ">>>", 3) != 0) continue;
	if( --num_topic == 0) break;
	}
#endif
if(num_topic != 0) return;
#ifdef WIN_MAC
q = strstr(p, "\r>>>"); if(q == NULL) q = p + strlen(p);
memcpy(txt, p, q - p);
txt[q - p] = 0;
#else
p = txt;
while(fgets(ligne, sizeof(ligne), help_file) != NULL) {
	if(strncmp(ligne, ">>>", 3) == 0) break;
	strcpy(p, ligne); p += strlen(ligne);
	}
#endif
AppendText(help_doc, txt, NULL, NULL, p_win_data->police);
Show(help_doc);
}


void help_ok_callback(ButtoN bouton)
{
Hide( ParentWindow(bouton) );
}


#ifdef WIN_MAC

void load_help_topics(char *p, LisT topic_list)
{
char *q, topic[50];
int l;
while( p != NULL) {
	if(strncmp(p, ">>>", 3) == 0) {
		q = strchr(p, '\r');
		memcpy(topic, p + 3, q - p - 3);
		topic[q - p - 3] = 0;
		ListItem(topic_list, topic);
		}
	p = strchr(p, '\r');
	if(p != NULL) p++;
	continue;
	}
}

#else

void load_help_topics(FILE *fich, LisT topic_list)
{
char ligne[100], topic[50];
int l;

rewind(fich);
while(fgets(ligne, sizeof(ligne), fich) != NULL) {
	if(strncmp(ligne, ">>>", 3) != 0) continue;
	strcpy(topic, ligne + 3);
	l = strlen(topic) - 1;
	topic[l] = 0;
	ListItem(topic_list, topic);
	}
}

#endif


void subtree_up_callback(ButtoN bouton)
{
if(subtree_center == NULL) return;
subtree_center = subtree_ascend;
if(subtree_center != NULL) {
	subtree_ascend = subtree_center->v3;
	if(subtree_ascend == NULL) {
		SetValue(fd_nj_plot->choix_group, show_tree);
		Disable(fd_nj_plot->subtree_up_button);
		choix = show_tree;
		show_tree_callback();
		return;
		}
	subtree_notu = calc_n_desc(subtree_center) - 1;
	scrollnotu(subtree_notu);
	}
need_runtree = TRUE;
tree_draw_proc(fd_nj_plot->tree_plot);
}


void process_keys(Nlm_Char key)
{
#ifdef WIN_MAC
	/* combinaison command-lettre */
#define test_key(a) Nlm_cmmdKey && Nlm_currentKey == a
#else
	/* combinaison control-lettre */
#define test_key(a)  key == a - 96
#endif
if( test_key('q') ) 
	exit(0);
else if( test_key('o') )
	open_callback(fd_nj_plot->open_button);
else if( test_key('s') )
	save_plot_callback(fd_nj_plot->save_plot_button);
else if( test_key('f') )
	search_callback(fd_nj_plot->SearchItem);
else if( test_key('a') ) {
	if(Enabled(fd_nj_plot->AgainItem) ) 
		search_callback(fd_nj_plot->AgainItem);
	}
#ifdef MICRO
else if( test_key('c') )
	copy_plot(NULL);
else if( test_key('p') )
	print_plot(NULL);
#endif
else if( test_key('v') )
	paste_tree(NULL);
return;
}


static void quitproc(WindoW i)
{
exit(0);
}


void win_resize_proc(WindoW win)
{
RecT *old_rect, rect;
int change_x, change_y, old_height, new_height;

old_rect = (RecT *)GetWindowExtra(win);
if(old_rect->left == -1) return;
ObjectRect(win, &rect);
old_height = old_rect->bottom - old_rect->top;
new_height = rect.bottom - rect.top;
change_y = (new_height - old_height);
old_height = old_rect->right - old_rect->left;
new_height = rect.right - rect.left;
change_x = (new_height - old_height);
if(change_x != 0 || change_y != 0) {
	change_panel_size(fd_nj_plot->tree_plot, change_x, change_y);
	*old_rect = rect;
	}
}


void change_panel_size(PaneL panel, int change_x, int change_y)
{
RecT r, r_group;
GrouP group;
Nlm_Boolean in_group;
Nlm_BaR bar;
int visible;
static int recursive = FALSE;

GetPosition(panel, &r);
#ifdef WIN_MOTIF
if(recursive) return;
recursive = TRUE;
bar = Nlm_GetSlateVScrollBar( (Nlm_SlatE)panel);
visible = Nlm_Visible(bar);
group = Parent(panel);
in_group = (group != ParentWindow(panel));
if(in_group) {
	GetPosition(group, &r_group);
	Hide(group);
	}
else
	Hide(panel); 
#endif
r.bottom += change_y;
r.right += change_x;
SetPosition(panel, &r);
#ifdef WIN_MOTIF
if(in_group) {
	r_group.bottom += change_y;
	r_group.right += change_x;
	SetPosition( group, &r_group);
	Show(group);
	}
else	Show(panel);
if( !visible) Nlm_Hide(bar); 
recursive = FALSE;
#endif
}


void font_callback(ChoicE choix)
{
prepare_fonts();
need_runtree = TRUE;
tree_draw_proc(fd_nj_plot->tree_plot);
}


void prepare_fonts(void)
{
Nlm_Boolean use_bold, use_italic;
Int2 font_num;
char font_full_name[40];

/* prepare vibrant font */
use_bold = GetStatus(fd_nj_plot->bold_item);
use_italic = GetStatus(fd_nj_plot->italic_item);
font_num = GetValue(fd_nj_plot->choix_font);
current_font_size_rank = GetValue(fd_nj_plot->choix_taille);
sprintf(font_full_name, "%s,%d", list_font_name[font_num], 
	list_font_size[current_font_size_rank]);
if(use_bold || use_italic) strcat(font_full_name, ",");
if(use_bold) strcat(font_full_name, "b");
if(use_italic) strcat(font_full_name, "i");
current_font = ParseFont(font_full_name);
prepare_ps_font(font_num, use_bold, use_italic);
}


/* callbacks for form nj_plot */
void open_callback(IteM ob)
{
int reponse;
char title[200];

reponse = MG_GetInputFName(fname, sizeof(fname));
if(!reponse) {
	tree_draw_proc(fd_nj_plot->tree_plot);
	return;
	}
if(notu != 0) free_tree();
notu = 0;
init_tree( extract_filename(fname) );
}


void paper_callback(ChoicE ob)
{
paper_choice = (paper_item) GetValue(ob);
}


void save_plot_callback(IteM ob)
{
double currx, curry;
char *p;
if(notu==0)return;
strcpy(plotfilename,fname);
p = strchr(extract_filename(plotfilename), '.'); 
if(p == NULL) p = plotfilename + strlen(plotfilename);
#ifdef WIN_MAC
strcpy(p,".pict");
#else
strcpy(p,".ps");
#endif
if(ob != NULL) {
	if( ! MG_GetOutputFName(plotfilename, sizeof(plotfilename), 
		extract_filename(plotfilename) ) ) return;
	}

#ifdef WIN_MAC
pict_plot();
#else
postscript_plot();
#endif
}


void save_tree_callback(IteM ob)
{
char *p, *arbre;
FILE *out;

if(notu==0)return;
/* build out file name */
strcpy(rooted_fname, fname);
#ifndef WIN_MSWIN
/* ajouter _root. dans le nom */
#define AJOUT "_root."
if(strstr(rooted_fname, AJOUT) == NULL) {
	p = strchr(rooted_fname, '.'); 
	if(p == NULL) p = rooted_fname + strlen(rooted_fname);
	strcpy(p, AJOUT);
	}
#undef AJOUT
p = strchr(fname, '.');
if(p == NULL)
	strcat(rooted_fname, "ph");
else
	strcpy( strchr(rooted_fname,'.'), p);
#else
/* remplacer extension par .phr */
p = strchr(rooted_fname, '.');
if(p == NULL) p = rooted_fname + strlen(rooted_fname);
strcpy( p, ".phr");
#endif
if( !MG_GetOutputFName(rooted_fname, sizeof(rooted_fname), extract_filename(rooted_fname) ) ) 
		return;
arbre = ecrit_arbre_parenth(racine);
if(arbre == NULL) {
	err_message("Sorry, not enough memory");
	return;
	}
out = fopen(rooted_fname,"w");
if(out != NULL) {
	fputs(arbre, out); putc('\n', out);
	fclose(out);
	}
else	{
	char mess[250];
	sprintf(mess, "Sorry, cannot write to file %s", 
		extract_filename(rooted_fname) );
	err_message(mess);
	}
free(arbre);
}


void toggle_branch_callback(ButtoN ob)
{
	plot_br_l = GetStatus(ob);
	need_runtree = TRUE;
	tree_draw_proc(fd_nj_plot->tree_plot);
}


void toggle_bootstrap_callback(ButtoN ob)
{
show_bootstrap = !show_bootstrap;
need_runtree = TRUE;
tree_draw_proc(fd_nj_plot->tree_plot);
}


void operation_callback(GrouP group)
{
choix = (enum operations) GetValue(group);
if(choix == permutation)
	swap_callback();
else if(choix == depl_racine)
	new_outgroup_callback();
else if(choix == subtree)
	subtree_callback();
else {
	show_tree_callback();
	}
}

void subtree_callback(void)
{
	choix = subtree;
	Disable(fd_nj_plot->save_tree_button);
	Disable(fd_nj_plot->open_button);
	Enable(fd_nj_plot->subtree_up_button);
	if(has_br_length) {
		SetStatus(fd_nj_plot->branch_length_button, FALSE);
		Disable(fd_nj_plot->branch_length_button);
		plot_br_l=0;
		}
	SetStatus(fd_nj_plot->bootstrap_button, FALSE);
	show_bootstrap = FALSE;
	Disable(fd_nj_plot->bootstrap_button);
	Disable(fd_nj_plot->reroot_button);
need_runtree = TRUE;
tree_draw_proc(fd_nj_plot->tree_plot);
}

void swap_callback(void)
{
	choix = permutation;
	Disable(fd_nj_plot->save_plot_button);
	Disable(fd_nj_plot->save_tree_button);
	Disable(fd_nj_plot->open_button);
	if(has_br_length) {
		SetStatus(fd_nj_plot->branch_length_button, FALSE);
		Disable(fd_nj_plot->branch_length_button);
		plot_br_l=0;
		}
	SetStatus(fd_nj_plot->bootstrap_button, FALSE);
	show_bootstrap = FALSE;
	Disable(fd_nj_plot->bootstrap_button);
need_runtree = TRUE;
tree_draw_proc(fd_nj_plot->tree_plot);
}

void new_outgroup_callback(void)
{
	choix = depl_racine;
	Disable(fd_nj_plot->save_plot_button);
	Disable(fd_nj_plot->save_tree_button);
	Disable(fd_nj_plot->open_button);
	if(has_br_length) {
		SetStatus(fd_nj_plot->branch_length_button, FALSE);
		Disable(fd_nj_plot->branch_length_button);
		plot_br_l=0;
		}
	SetStatus(fd_nj_plot->bootstrap_button, FALSE);
	show_bootstrap = FALSE;
	Disable(fd_nj_plot->bootstrap_button);
need_runtree = TRUE;
tree_draw_proc(fd_nj_plot->tree_plot);
}


void show_tree_callback(void)
{
	choix = show_tree;
	Enable(fd_nj_plot->save_plot_button);
	Enable(fd_nj_plot->save_tree_button);
	Enable(fd_nj_plot->open_button);
	Disable(fd_nj_plot->subtree_up_button);
	if(has_br_length) {
		Enable(fd_nj_plot->branch_length_button);
		}
	if(has_internal) {
		Enable(fd_nj_plot->bootstrap_button);
		}
	Enable(fd_nj_plot->reroot_button);
need_runtree = TRUE;
subtree_notu = notu;
if(subtree_center != NULL) scrollnotu(notu);
subtree_center = NULL;
tree_draw_proc(fd_nj_plot->tree_plot);
}


void exit_callback(IteM ob)
{
exit(0);
}


void tree_draw_proc(PaneL panel)
{
RecT ob_rect;
Int2 width, height;
static Int2 previous_h = 0, previous_w = 0;
Nlm_BaR bar;
int hidebar;

WatchCursor();
bar = Nlm_GetSlateVScrollBar((Nlm_SlatE)panel);
Nlm_GetPanelExtra(panel, &hidebar);
if(hidebar && Nlm_Visible(bar)) Nlm_Hide(bar);
else if( !hidebar && !Nlm_Visible(bar)) Nlm_Show(bar);
Select (panel);
ObjectRect(panel, &ob_rect);
EraseRect(&ob_rect);
FrameRect(&ob_rect);
Nlm_ClipRect(&ob_rect);
	if(notu == 0 || strlen(fd_nj_plot->tree_name) == 0) {
		ArrowCursor();
		return;
		}
	width = ob_rect.right - ob_rect.left;
	height = ob_rect.bottom - ob_rect.top;
	SelectFont(current_font);
	Black();
	if( need_runtree || /* recalcul des positions des # si window re-size */
		(choix != show_tree && (height != previous_h || width != previous_w)))
			runtree();
	need_runtree = FALSE;
	swap = 0;
	do_plot();
	previous_w = width;
	previous_h = height;
Nlm_ResetClip();
ArrowCursor();
}


void tree_click_proc(PaneL panel, PoinT click)
{
	double eps, x, y;
	int i, found=0, node_num;
	RecT ob_rect;
	Int2 panel_x, panel_y, panel_h;
	if(notu==0) return;
	if(choix != permutation && choix != depl_racine && choix != subtree) 
		return;
	ObjectRect(panel, &ob_rect);
	panel_x = ob_rect.left; panel_y = ob_rect.top; panel_h = ob_rect.bottom - ob_rect.top;
	eps = char_height/2;
	x = click.x - panel_x;
	click.y -= panel_y;
	y = panel_h - click.y;
	for(i=0; i<=totpoints; i++ ) {
		if(fabs(x - (points+i)->x) > eps)
			 continue;
		if(fabs(y - (points+i)->y) > eps)
			 continue;
		found=1;
		break;
		}
	if(!found)return;
	node_num=(points+i)->number;
	if(choix==depl_racine) {
		if(node_num >= 1 && node_num <= 2*notu && 
						node_num-1 != root_num) {
			root_num = node_num-1;
			removeroot();
			}
		}
	else if(choix==permutation) {
		if(node_num >= notu+2 && node_num <= 2*notu+1) 
			swap= node_num-1;
		else
			swap=0;
		}
	else if(choix==subtree) {
		if(node_num < notu+2 || node_num > 2*notu+1) return;
		subtree_center = tabtax[node_num - 1];
		subtree_ascend = subtree_center->v3;
		subtree_notu = calc_n_desc(subtree_center) - 1;
		scrollnotu(subtree_notu);
		choix = show_tree;
		SetValue(fd_nj_plot->choix_group, 0);
		Nlm_InvalObject(fd_nj_plot->choix_group);
		if(has_br_length) {
			Enable(fd_nj_plot->branch_length_button);
			Nlm_InvalObject(fd_nj_plot->branch_length_button);
			}
		if(has_internal) {
			Enable(fd_nj_plot->bootstrap_button);
			Nlm_InvalObject(fd_nj_plot->bootstrap_button);
			}
		}
	need_runtree = TRUE;
	tree_draw_proc(panel);
}


void paste_tree(IteM unused)
{
FILE *tmp;
char *buff, *tmpfname;
char pasted_title[] = "pasted tree";
if(notu != 0) {
	err_message("Do \"Edit:clear\" before pasting tree data");
	return;
	}

#ifdef WIN_MAC
Nlm_textScrapFull = FALSE; /* arrange un bug de Nlm_ClipboardHasString */
#endif
if( ! Nlm_ClipboardHasString() ) {
	err_message("No text describing tree present in clipboard");
	return;
	}
buff = Nlm_ClipboardToString();
tmpfname = tmpnam(NULL);
tmp=fopen(tmpfname,"w");
fwrite(buff, 1, strlen(buff), tmp);
fclose(tmp);
strcpy(fname, tmpfname);
init_tree(pasted_title);
remove(tmpfname);
if(notu > 0) {
	strcpy(fname, pasted_title);
	}
else	fname[0] = 0;
}



void clear_tree(IteM unused)
{
if(notu != 0) free_tree();
notu = 0;
SetTitle(fd_nj_plot->nj_plot, "njplot");
tree_draw_proc(fd_nj_plot->tree_plot);
}


void change_page_count(ButtoN item)
{
Nlm_Boolean *pdone;

pdone = (Nlm_Boolean *)GetWindowExtra( Nlm_ParentWindow(item) );
*pdone = TRUE;
}


void page_count_callback(IteM item)
{
static WindoW win;
static Nlm_Boolean done;
static TexT select_box;
char select[100];
static int first = TRUE;
	if(first) {
		win = FixedWindow(-50,-50,-5,-5,"Page count:",NULL);
		sprintf(select, "%d", page_count);
		select_box=DialogText(win,select,8,NULL);
		Advance(win);
		DefaultButton(win,"ok",change_page_count);
		SetWindowExtra(win, &done, NULL);
		first = FALSE;
		}
	Show(win);
	Select(win); Select(select_box);
	done=FALSE;
	while(! done) {
		myWaitAndProcessNextEvent();
		}
	GetTitle(select_box, select, sizeof(select));
	sscanf(select, "%d", &page_count);
	Hide(win);
#ifndef WIN_MSWIN /* changement titre menu item impossible sur PC ! */
	sprintf(select,"Page count (%d)", page_count);
	Nlm_SetTitle(item, select);
#endif	
	Select(fd_nj_plot->nj_plot);
}




void string_callback(ButtoN item)
{
Nlm_Boolean *pdone;

pdone = (Nlm_Boolean *)GetWindowExtra( Nlm_ParentWindow(item) );
*pdone = TRUE;
}


void search_callback(IteM item)
{
static WindoW win;
static Nlm_Boolean search_done;
static TexT select_box;
static char select[500];
char aux[500];
int num, trouve;
static int first = TRUE;

if(item == fd_nj_plot->SearchItem) {
	if(first) {
		win = FixedWindow(-50,-50,-5,-5,"Name searched:",NULL);
		select_box=DialogText(win,"",15,NULL);
		Advance(win);
		DefaultButton(win,"ok",string_callback);
		SetWindowExtra(win, &search_done, NULL);
		first = FALSE;
		}
	Show(win);
	Select(win); Select(select_box);
	search_done=FALSE;
	while(! search_done) {
		myWaitAndProcessNextEvent();
		}
	GetTitle(select_box, select, sizeof(select));
	majuscules(select);
	Hide(win);
/*	Enable(fd_nj_plot->AgainItem); */
	Select(fd_nj_plot->nj_plot);
	}
else if(item != fd_nj_plot->AgainItem) { /* happens only when called after start  */
	strcpy(select, (char *)item);
	majuscules(select);
	}
if(strlen(select) == 0) return;
trouve = FALSE;
for(num = 0; num <= totnoms; num++) {
	strcpy(aux, (noms+num)->nom);
	majuscules(aux);
	if(strstr( aux, select) != NULL) {
		(noms+num)->disp_option = 'r';
		trouve = TRUE;
		}
	}
if(trouve) {
	tree_draw_proc(fd_nj_plot->tree_plot);
	}
}


void init_tree(char *titre)
{
char *pname;
char titre_l[300];

choix = show_tree;
WatchCursor();
/* read tree file */
if( (pname=preptree(fname)) != NULL ) {
	char *message;
	size_t taille;
	fd_nj_plot->tree_name[0] = 0;
	notu = 0;
	SetTitle(fd_nj_plot->nj_plot, "njplot");
	tree_draw_proc(fd_nj_plot->tree_plot);
	taille = strlen(pname) + strlen(fname) + 3;
	message = check_alloc(1, taille);
	strcpy(message, pname);
	strcat(message, " ");
	strcat(message, extract_filename(fname) );
	err_message(message);
	free(message);
	return;
	}
maxy = 1000.;
deltay = maxy/notu;
fd_nj_plot->tree_name = fname;
need_runtree = TRUE;
subtree_center = NULL;

strcpy(titre_l, titre );
if(notu > 0) sprintf(titre_l + strlen(titre_l), " (%d tips)", notu + 1);
SetTitle(fd_nj_plot->nj_plot, titre_l);

tree_draw_proc(fd_nj_plot->tree_plot);
if(has_br_length) {
	Enable(fd_nj_plot->branch_length_button);
	}
else
	Disable(fd_nj_plot->branch_length_button);
if(has_internal) {
	Enable(fd_nj_plot->bootstrap_button);
	}
else
	Disable(fd_nj_plot->bootstrap_button);
}


/* fenetre message d'alerte.
Operation Select dans la boucle des evenements qui ramene toujours la fenetre
au premier plan. Lent, mais interessant.
*/
static void alert_ok_action(ButtoN bouton)
{
Nlm_Boolean *alert_done;
alert_done = (Nlm_Boolean *)GetWindowExtra(Parent(bouton));
*alert_done = TRUE;
}


void err_message(const char *in)
{
Nlm_WindoW alwin;
Nlm_Boolean alert_done;
Nlm_TexT t;
int nl;
char *p, *q, *r;
static char texte[1000];
int w, width;

#ifdef WIN_MAC
alwin = Nlm_FixedWindow(-50,-50,-5,-5,"MESSAGE",NULL); // Bug: ModalWindow crashes when return key hit
#else
alwin = Nlm_ModalWindow(-50,-50,-5,-5,NULL);
Nlm_SetTitle(alwin, "Message");
#endif
Nlm_SelectFont (Nlm_systemFont);
nl = 0; p = (char *)in; r = texte; width = 0;
while(TRUE)	{
	q = strchr(p,'\n');
	if(q == NULL) q = p + strlen(p);
	nl++; 
	if( (r - texte) + (q - p) + 2 >= sizeof(texte) ) break;
	if(q > p) {
		memcpy(r, p, q - p); 
		w = Nlm_TextWidth(p, q - p);
		if(w > width) width = w;
		}
	r += q - p;
#ifdef WIN_MSWIN
	*(r++) = '\r';
#endif
#ifdef WIN_MAC
	*(r++) = '\r';
#else
	*(r++) = '\n';
#endif
	*r = 0;
	q = strchr(p,'\n');
	if(q == NULL) break;
	p = q + 1; 
	}

width = ( width + 20 ) / Nlm_stdCharWidth;
t = Nlm_ScrollText(alwin, width, nl , Nlm_systemFont, TRUE, NULL);

Nlm_SetTitle(t, texte);
Nlm_Break(alwin);
Nlm_DefaultButton(alwin,"ok",alert_ok_action);
Nlm_Show(alwin);
alert_done=FALSE;
Nlm_SetWindowExtra(alwin, &alert_done, NULL);
Nlm_ArrowCursor();
while(! alert_done) {
#ifdef WIN_MAC
	if(! Nlm_InFront(alwin) ) 
		Nlm_Select(alwin); // to emulate modal window on mac
#endif
	myWaitAndProcessNextEvent();
	}
Nlm_Remove(alwin);
}

#else /* NO_GUI */

void err_message(const char *in)
{
fprintf(stderr, "%s\n", in);
}

#endif /* NO_GUI */


void prepare_ps_font(int font_num, int use_bold, int use_italic)
{
/* prepare postscript name of font */
strcpy(current_ps_font, list_ps_font_name[font_num]);
current_ps_font[0] = toupper(current_ps_font[0]);
if( use_bold )
	strcat(current_ps_font,"-Bold");
else if( use_italic )  {
	if(font_num == times) 
		strcat(current_ps_font,"-Italic");
	else
		strcat(current_ps_font,"-Oblique");
	}
else if(font_num == times) 
	strcat(current_ps_font,"-Roman");
}


int direct_ps_plot(void)
{
char *p;

choix = show_tree;
/* read tree file */
if( (p = preptree(fname) ) != NULL ) {
	fprintf(stderr, "%s\n", p);
	return 1;
	}
strcpy(plotfilename, fname);
p = strchr(plotfilename, '.'); 
if(p == NULL) p = plotfilename + strlen(plotfilename);
#ifdef TTY
strcpy(p, ".txt");
#else
strcpy(p, ".ps");
#endif
#ifdef TTY
#elif defined(NO_GUI)
	prepare_ps_font(courier, FALSE, FALSE);
#else
	prepare_ps_font(times, FALSE, FALSE);
#endif
maxy = 1000.;
deltay = maxy/notu;
ps_plot = TRUE;
runtree();
#ifdef ADDROOT
/* calcul du nombre de feuilles (tres obligatoire !!!) */
calc_otu(NULL, racine);
/*  calcul de son format phylip */
p = ecrit_arbre_parenth(racine);
if(p == NULL) {
	fprintf(stderr, "Sorry, not enough memory\n");
	exit(1);
	}
/* ecriture de son format phylip */
printf("%s\n", p);
#elif defined(TTY)
tty_plot();
#else
postscript_plot();
#endif
return 0;
}


int calc_otu(struct noeud *pere, struct noeud *centre)
{
struct noeud *gauche, *droite;
double bg, bd, bpere;

if(centre == NULL) return 0;

/* orienter le noeud centre de maniere standard: centre->v3=pere */
if( centre->v1 == pere ) {
	gauche =centre->v2; droite = centre->v3;
	bg = centre->l2; bd = centre->l3; bpere = centre->l1;
	}
else if( centre->v2 == pere ) {
	gauche =centre->v1; droite = centre->v3;
	bg = centre->l1; bd = centre->l3; bpere = centre->l2;
	}
else	{
	gauche =centre->v1; droite = centre->v2;
	bg = centre->l1; bd = centre->l2; bpere = centre->l3;
	}
centre->v3=pere; centre->v1=gauche; centre->v2=droite;
centre->l3=bpere; centre->l1=bg; centre->l2=bd;

return calc_otu(centre, centre->v1) + calc_otu(centre, centre->v2);
}


void remove_arg(int target, int *argc, char *argv[])
{
int num;
for(num = target; num < *argc - 1; num++) 
	argv[num] = argv[num+1];
(*argc)--;
}


void process_args(int *argc, char *argv[], char **pred)
{
int num, taille;

for(num = 1; num < *argc; num++) {
	if( strncmp(argv[num], "-h", 2) == 0 ){
		char message[1000];
		sprintf(message,
		"Usage: %s [-options] [tree_file_name]\n"
"where options are:\n"
"-h             print out this message\n"
#ifdef ADDROOT
"-outotu name   name of species to use as outgroup for rooting\n"
#elif defined(TTY)
"-txtw n        width of output text (default = 120)\n"
#else
"-us            postcript tree file prepared for US Letter paper size\n"
#ifndef NO_GUI
"-psonly        no window interface, just write the postscript tree file\n"
"-red xxxx      highlight in red all taxon names containing string xxxx\n"
#endif
"-pc n          number of pages for PostScript output\n"
"-size n        font size n used for taxon names\n"
"-lengths       show branch lengths if they appear in tree file\n"
"-boot          show bootstrap values if they appear in tree file\n"
"-pssize        size of page for postscript expressed as WIDTHxHEIGHT\n"
"-notitle       don't include title in postscript output\n"
#endif
"\n"
"and where tree_file_name is the name of a tree file in the Newick parentheses format\n" 
, extract_filename(argv[0]) );

#ifdef MICRO
		err_message(
#else
		fprintf(stderr,
#endif
		message);
		exit(0);
		}
	}

for(num = 1; num < *argc; num++) {
	if( strncmp(argv[num], "-us", 3) == 0) {
		paper_choice = LETTER;
		remove_arg(num, argc, argv);
		break;
		}
	}

for(num = 1; num < *argc; num++) {
	if( strncmp(argv[num], "-size", 5) == 0) {
		const int maxi = sizeof(list_font_size) / sizeof(int);
		taille = 12;
		if(num + 1 < *argc) {
			sscanf(argv[num + 1], "%d", &taille);
			remove_arg(num + 1, argc, argv);
			}
		remove_arg(num, argc, argv);
		for(num = 0; num < maxi; num++) 
			if(taille <= list_font_size[num]) break;
		if(num >= maxi) 
			num = maxi - 1;
		current_font_size_rank = num;
		break;
		}
	}
for(num = 1; num < *argc; num++) {
	if( strncmp(argv[num], "-pc", 3) == 0) {
		if(num + 1 < *argc) {
			sscanf(argv[num + 1], "%d", &page_count);
			remove_arg(num + 1, argc, argv);
			}
		remove_arg(num, argc, argv);
		break;
		}
	}
for(num = 1; num < *argc; num++) {
	if( strncmp(argv[num], "-psonly", 7) == 0) {
		ps_plot_only = TRUE;
		remove_arg(num, argc, argv);
		break;
		}
	}
for(num = 1; num < *argc; num++) {
	if( strncmp(argv[num], "-lengths", 8) == 0) {
		plot_br_l = TRUE;
		remove_arg(num, argc, argv);
		break;
		}
	}
for(num = 1; num < *argc; num++) {
	if( strncmp(argv[num], "-boot", 5) == 0) {
		show_bootstrap = TRUE;
		remove_arg(num, argc, argv);
		break;
		}
	}
for(num = 1; num < *argc; num++) {
	if( strncmp(argv[num], "-pssize", 7) == 0) {
		if(num + 1 < *argc) {
			sscanf(argv[num + 1], "%dx%d", &ps_width, &ps_height);
			remove_arg(num + 1, argc, argv);
			}
		remove_arg(num, argc, argv);
		break;
		}
	}
for(num = 1; num < *argc; num++) {
	if( strncmp(argv[num], "-notitle", 8) == 0) {
		no_title = TRUE;
		remove_arg(num, argc, argv);
		break;
		}
	}
*pred = NULL;
#ifndef NO_GUI
for(num = 1; num < *argc; num++) {
	if( strncmp(argv[num], "-red", 4) == 0) {
		if(num + 1 < *argc) {
			*pred = argv[num+1];
			remove_arg(num + 1, argc, argv);
			}
		remove_arg(num, argc, argv);
		break;
		}
	}
#elif defined(TTY)
physx = 120; /* default width of text output */
for(num = 1; num < *argc; num++) {
	if( strncmp(argv[num], "-txtw", 5) == 0) {
		if(num + 1 < *argc) {
			sscanf(argv[num+1], "%d", &physx);
			remove_arg(num + 1, argc, argv);
			}
		remove_arg(num, argc, argv);
		break;
		}
	}
#elif defined(ADDROOT)
for(num = 1; num < *argc; num++) {
	if(strcmp(argv[num], "-outotu") == 0) {
		if(num + 1 < *argc) {
			outotu = argv[num+1];
			remove_arg(num + 1, argc, argv);
			}
		remove_arg(num, argc, argv);
		break;
		}
	}
#endif
}


double length_log_phys(double p)
{
return p * tek_dx;
}

double length_phys_log(double p)
{
return  p / tek_dx;
}


double arrondi_echelle(double x)
{ /* arrondi x a une valeur 1, 2, 5 pour echelle */
double l, n;
int r;
static int corresp[] = {1,1,2,2,5,5,5,10,10,10,10,10};
                     /* 0,1,2,3,4,5,6, 7, 8, 9,10,11 */
l = log10(x);
n = floor(l);
l = x * pow(10., -n);  /* 10. plutot que 10 necessaire pour alpha! */
r = myrint(l); r = corresp[r];
return r * pow(10., n);  /* 10. plutot que 10 necessaire pour alpha! */
}


void scale_window(double lxmin, double lxmax, double lymin, double lymax,
 double pxmin, double pxmax, double pymin, double pymax)
/* to scale the plot window for logical coords l... in physical coords p... */
{
tek_xmin=lxmin; tek_xmax=lxmax; tek_ymin=lymin; tek_ymax=lymax;
tek_minx=pxmin; tek_maxx=pxmax; tek_miny=pymin; tek_maxy=pymax;
tek_dx = (tek_maxx-tek_minx)/(tek_xmax-tek_xmin);
tek_dy = (tek_maxy-tek_miny)/(tek_ymax-tek_ymin);
}


void ch_echelle(double lx, double ly, double *px, double *py)
/* conversion coord logique lx,ly en coord physique px,py */
{
*px = (lx-tek_xmin)*tek_dx + tek_minx;
*py = (ly-tek_ymin)*tek_dy + tek_miny;
}


void convert_mem_point(void)
{
int i;
if(converted) return;
for (i=0; i<=totpoints; i++) {
   	ch_echelle( (points+i)->x, (points+i)->y, &((points+i)->x), 
			&((points+i)->y) );
	(points+i)->x += char_height / 6.;
	}
converted = TRUE;
}


double calc_echelle(double larg)
{ /* rend taille logique pour echelle optimale */
double log_val, phys_val;
phys_val = larg/10;
log_val = length_phys_log(phys_val);
log_val = arrondi_echelle(log_val);
return log_val;
}


void draw_scale(void)
{
char ech_name[20];
double phys_w, y, xd, xf, lc;
double log_val;

log_val = calc_echelle(physx);
phys_w = myrint(length_log_phys(log_val));
y = physy - 1.35 * char_height;
xf = physx * 0.95;
xd = xf - phys_w;
sprintf(ech_name, "%.1g", log_val);
lc = calc_text_size(ech_name, NULL, NULL);
dir_moveto(xd, y);
dir_lineto(xf, y);
#ifndef TTY
dir_moveto(xd, y - char_height/3.);
dir_lineto(xd, y + char_height/3.);
dir_moveto(xf, y - char_height/3.);
dir_lineto(xf, y + char_height/3.);
#endif
dir_moveto( (xd + xf)/2 - lc/2, 
	y + char_height/6. ); 
plotstring(ech_name);
}


#ifndef NO_GUI
void scrollcallback(Nlm_BaR b, Nlm_SlatE g, Nlm_Int2 after, Nlm_Int2 before)
{
tree_draw_proc(fd_nj_plot->tree_plot);
}


void zoomcallback(Nlm_SwitcH s, Nlm_Int2 after, Nlm_Int2 before)
{
char aux[50];
int max, hide;

if(after == 1) fd_nj_plot->zoomvalue = 1.;
else if(after > before) fd_nj_plot->zoomvalue *= 1.4;
else if(after < before) fd_nj_plot->zoomvalue /= 1.4;
sprintf(aux, "%d%%", (int)(100. * fd_nj_plot->zoomvalue + 0.5) );
Nlm_SetTitle(fd_nj_plot->zoomprompt, aux);
Nlm_InvalObject(fd_nj_plot->zoomprompt);
scrollnotu(subtree_notu);
hide = (after == 1);
Nlm_SetPanelExtra(fd_nj_plot->tree_plot, &hide);
tree_draw_proc(fd_nj_plot->tree_plot);
}


void scrollnotu(int count)
{
Nlm_BaR obj; int lines;
float old;

	obj = Nlm_GetSlateVScrollBar( (SlatE)fd_nj_plot->tree_plot );
	old = Nlm_GetBarValue(obj) / (float)Nlm_GetBarMax(obj);
	lines = (count + 1) * (1 - 1/fd_nj_plot->zoomvalue) + 0.5; 
	if(lines < 3) lines = 3;
	Nlm_CorrectBarMax(obj, lines);
	Nlm_CorrectBarValue(obj, old * lines);
	Nlm_CorrectBarPage(obj, lines/fd_nj_plot->zoomvalue, lines/fd_nj_plot->zoomvalue);
}

#endif

void do_plot(void)
{
struct trait *p;
int num, erreur = FALSE;
int offset;
double factor, value, m = 1, v = 0, zoom = 1;

#ifndef NO_GUI
/* size of the tree plot window  */
if( ! (ps_plot || doing_print) ) {
		Nlm_RecT ob_rect; Nlm_BaR bar;
		Nlm_ObjectRect(fd_nj_plot->tree_plot, &ob_rect);
		bar = Nlm_GetSlateVScrollBar((Nlm_SlatE)fd_nj_plot->tree_plot);
		v = (double)Nlm_GetBarValue(bar);
		m = (double)Nlm_GetBarMax(bar);
		physx = ob_rect.right - ob_rect.left;
		physy = (ob_rect.bottom - ob_rect.top);
		zoom = fd_nj_plot->zoomvalue;
	}
#endif

factor = 1e99;
for(num=0; num<=notu; num++) {
	if(profs[num] == 0) continue;
	if(profs[num] <= 0) continue; /* new */
	value = (physx - widnames[num] - 2 * ascent) / profs[num];
	if(value < factor) factor = value;
	}

physx_min  = ascent;
physx_corr = maxx * factor + physx_min;
if(physx_corr <= physx_min + 1) {
	physx_corr = physx_min + 1;
	}
physy_corr = physy - 2*char_height;
physy_min  = char_height;
#ifdef TTY
physy_corr = physy - 3;
physy_min  = 0;
#endif
/* scale_window(0, maxx, 0, maxy, 
	physx_min, physx_corr, physy_min, physy_corr); */
scale_window(0, maxx, 
	((m-v)/m) * (maxy - maxy/zoom), 
	((m-v)/m) * (maxy - maxy/zoom) + maxy/zoom, 
	physx_min, physx_corr, physy_min, physy_corr);

for(num=0; num<=totnoms; num++) {
	mydrawstring( (noms+num)->x, (noms+num)->y , (noms+num)->nom,
			(noms+num)->disp_option );
	}
for(num=0; num<=tottraits; num++) {
	p= traits+num;
	moveto(p->xd,p->yd);
	lineto(p->xf,p->yf);
	}

if(has_br_length){
	/* echelle */
	draw_scale();
	}

convert_mem_point();
} /* end of do_plot */



int calc_text_size(char *text, int *pheight, int *pascent)
{
int ascent, descent, width;

if(ps_plot) {
	int font_size;
	font_size = list_font_size[current_font_size_rank];
#ifdef WIN_MAC
	{
	static FontInfo macfontinfo;
	width = TextWidth(text, 0, strlen(text));
	GetFontInfo(&macfontinfo);
	ascent = macfontinfo.ascent;
	descent = macfontinfo.descent;
	}
#elif defined(TTY)
	width = strlen(text);
	ascent = 1; descent = 0;
#elif defined(NO_GUI)
	width = strlen(text) * font_size * postscript_ratio + 0.5;
	ascent = font_size; descent = 0;
#else
/*
	faire calculer taille en mode ecran et appliquer ratio ecran->postscript
*/
	width = Nlm_TextWidth(text, strlen(text) ) * postscript_ratio + 0.5;
	ascent = font_size; descent = 0;
#endif
	}
#ifndef NO_GUI
else	{
#ifdef WIN_MSWIN
	HDC picHDC = Nlm_currentHDC;
	if(doing_copy) {
		Nlm_currentHDC = Nlm_GetPicWinHDC();
		Nlm_SelectFont(current_font);
		}
#endif
    ascent = Nlm_Ascent(); descent = Nlm_Descent();
	width = Nlm_TextWidth(text, strlen(text) );
#ifdef WIN_MSWIN
	if(doing_copy) Nlm_currentHDC = picHDC;
#endif
	}
#endif
if(pheight != NULL) *pheight = ascent + descent;
if(pascent != NULL) *pascent = ascent;
return width;
}


static int lastmovex, lastmovey;
#define bound_int2(i) ( i > INT2_MAX ? INT2_MAX : (i < INT2_MIN ? INT2_MIN : i) )

void plotstring(char *nom)
{
if(ps_plot) { 
#ifdef WIN_MAC
	static char copy[255];
	int l;
	l = strlen(nom);
	*copy = l;
	memcpy(copy + 1, nom, l + 1);
	DrawString( (ConstStr255Param) copy);
#elif defined(TTY)
	memcpy(tty_page[tty_y]+tty_x+1, nom, strlen(nom));
#else
	fprintf(plotfile,"(%s) show\n",nom);
#endif
	}
#ifndef NO_GUI
else 	{
#ifdef WIN_MSWIN
	int y; y = lastmovey - (doing_copy ? ascent : 0) ;
	y = bound_int2(y);
  	Nlm_PaintStringEx(nom, lastmovex, y );
#else
        Nlm_PaintString(nom);
#endif
	}
#endif
}


void dir_moveto(double x,double y)
/* move to physical coord x,y */
{
if(ps_plot) {
	int xi,yi;
	xi=x;
#ifdef WIN_MAC
	xi += margin;
	yi = physy - y + margin;
	MoveTo(xi, yi);
#elif defined(TTY)
	tty_x = x + 0.5; tty_y = y + 0.5;
#else
	yi=y;
	fprintf(plotfile,"%d %d moveto\n",xi,yi);
#endif
}
#ifndef NO_GUI
else	{
	if(doing_print) {
		lastmovex = myrint(x) + print_rect.left;
		lastmovey = print_rect.bottom - myrint(y);
		Nlm_MoveTo(lastmovex, bound_int2(lastmovey) );	
		}
	else {
		Nlm_RecT obrect;
		ObjectRect(fd_nj_plot->tree_plot, &obrect);
		lastmovex = myrint(x) + obrect.left;
		lastmovey = obrect.bottom - myrint(y);
		Nlm_MoveTo(lastmovex, lastmovey );	
		}
	}
#endif
}


void dir_lineto(double x,double y)
/* draw line from current pos to physical coord x,y */
{
int xi,yi;
char *p;
if(ps_plot) {
	xi=x;
#ifdef WIN_MAC
	xi += margin;
	yi = physy - y + margin;
	LineTo(xi, yi);
#elif defined(TTY)
	if(tty_y == (int)(y + 0.5)) {
		xi = x + 0.5;
		while(tty_x != xi) { 
			p = tty_page[tty_y] + tty_x; 
			*p = (*p == '|' ? '+' : '-'); 
			if(tty_x < xi) ++tty_x; else --tty_x;
			}
		p = tty_page[tty_y] + tty_x; 
		*p = (*p == '|' ? '+' : '-'); 
		}
	else {
		yi = y + 0.5;
		while(tty_y != yi) { 
			p = tty_page[tty_y] + tty_x; 
			*p = (*p == '-' ? '+' : '|'); 
			if(tty_y < yi) ++tty_y; else --tty_y;
			}
		p = tty_page[tty_y] + tty_x; 
		*p = (*p == '-' ? '+' : '|'); 
		}
#else
	yi=y;
	fprintf(plotfile,"%d %d lineto stroke\n",xi,yi);
#endif
	}
#ifndef NO_GUI
else	{
	if(doing_print) {
		xi =	myrint(x) + print_rect.left; yi = print_rect.bottom - myrint(y) ;
		yi = bound_int2(yi);
		Nlm_LineTo(	xi, yi );
		}
	else {
		Nlm_RecT obrect;
		ObjectRect(fd_nj_plot->tree_plot, &obrect);
		Nlm_LineTo(	myrint(x) + obrect.left, obrect.bottom - myrint(y) );
		}
	}
#endif
}



#ifdef WIN_MAC

int is_macosx(void)
{
long val;

Gestalt (gestaltSystemVersion, &val);
return (val >= 0x1000);
}


int crefpict(char *fname, PicHandle picture)
{
OSErr success = 0, err;
static char zeros[512];
Nlm_Boolean async = false;
static char strfname[255];
OSType f_type,f_creator;
static HParamBlockRec paramblock;
PicPtr picptr;
int size;

strcpy(strfname+1, fname);
*strfname = strlen(fname);
paramblock.fileParam.ioNamePtr = (StringPtr)&strfname;

plotfile = fopen(fname, "wb");
if( (plotfile == NULL) || (fwrite(zeros,1,512,plotfile) != 512) )
	return TRUE;
size = GetHandleSize((Handle)picture);
HLock( (Handle) picture);
picptr = *picture;
if( fwrite(picptr,1,size,plotfile) != size ) return TRUE;
HUnlock( (Handle) picture);
fclose(plotfile);
KillPicture(picture);

err = PBHGetFInfoAsync(&paramblock);
if( err != success ) return FALSE;
strncpy((char *)&f_type, "PICT", 4);
strncpy((char *)&f_creator, "ttxt", 4); /* creator = SimpleText */
paramblock.fileParam.ioFlFndrInfo.fdType = f_type;
paramblock.fileParam.ioFlFndrInfo.fdCreator = f_creator;
err = PBHSetFInfoAsync(&paramblock);
return FALSE;
}


void example_tree(IteM unused)
{
FILE *tmp;
char *tmpfname;
char ex_name[] = "example";
if(example == NULL) return;
if(notu > 0) free_tree();
notu = 0;
tmpfname = tmpnam(NULL);
tmp=fopen(tmpfname,"w");
fwrite(example, 1, strlen(example), tmp);
fclose(tmp);
strcpy(fname, tmpfname);
init_tree(ex_name);
remove(tmpfname);
strcpy(fname, ex_name);
}


void copy_plot(IteM unused)
{
unsigned long pictlen;
double currx, curry;
ScrapRef myscrapref;

if(notu == 0) return;
doing_copy = TRUE;
pict_plot();

ClearCurrentScrap();
HLock( (Handle) mypicture);
pictlen = GetHandleSize( (Handle)  mypicture);
GetCurrentScrap(&myscrapref);
PutScrapFlavor(myscrapref, 'PICT', 0, pictlen, (Ptr)*mypicture );
//PutScrap(pictlen, 'PICT', (Ptr)*mypicture);
HUnlock( (Handle) mypicture);
KillPicture(mypicture);
ps_plot = FALSE;
need_runtree = TRUE;
doing_copy = FALSE;
ArrowCursor();
}


void pict_plot(void)
{
char mess[250];
double currx, curry;
int erreur;
	int macfont;
	Style mystyle=normal;
	margin = 30;
	myrect.top=10;
	myrect.left=10;
	myrect.bottom=760 * page_count;
	myrect.right=500;
	ClipRect(&myrect);
	mypicture = OpenPicture(&myrect);
	PenNormal();
	macfont = GetValue(fd_nj_plot->choix_font);
	if(macfont == courier)
		macfont=22;
	else if(macfont == helvetica)
		macfont=21;
	else if(macfont == times)
		macfont=20;
	else
		macfont=22;
	TextFont(macfont);
	TextSize( list_font_size[current_font_size_rank] );
    if( GetStatus(fd_nj_plot->bold_item) )
    	mystyle += bold;
    if( GetStatus(fd_nj_plot->italic_item) )
    	mystyle += italic;
    TextFace(mystyle);
	physx = myrect.right-myrect.left - 2 * margin;
	physy = myrect.bottom-myrect.top - 2 * margin;

WatchCursor();
ps_plot = TRUE;
totnoms = tottraits = totpoints = -1;
end_br_length = br_length_txt;
currx = 0.;
maxx = 0.;
nexty = -deltay;
if(subtree_center == NULL)
	mem_plot(NULL, racine, currx, &curry);
else
	mem_plot(subtree_ascend, subtree_center, currx, &curry);
do_plot();
ps_plot = FALSE; /* pour eviter boucle infernale! */
	ClosePicture();
if(doing_copy) return;
	erreur = crefpict(plotfilename, mypicture);
	if( !erreur ) sprintf(mess, "Tree plot is now in file %s in PICT format", 
			extract_filename(plotfilename) );
	else sprintf(mess, "Error while writting to file %s", 
			extract_filename(plotfilename) );
	err_message(mess);

need_runtree = TRUE;
ArrowCursor();
}

#endif


#ifdef MICRO

#ifdef WIN_MSWIN
#define TITLEFONT "Times New Roman,10"
#else
#define TITLEFONT "Times,10"
#endif

void print_title(int x, int y, char *text, 
		Nlm_FonT title_font, int p, int totp)
{
static char ligne[200];
time_t heure;
int h;

time (&heure);
sprintf(ligne, "njplot    %s    %s", text, ctime(&heure) );
h = strlen(ligne) - 1; ligne[h] = 0;
if(totp > 1) sprintf(ligne + h, " Page %d of %d", p, totp);
Nlm_Black();
Nlm_SelectFont(title_font);
Nlm_PaintStringEx(ligne, x, y);
}


void print_plot(IteM unused)
{
double currx, curry;
Nlm_WindoW w;
int h, page, superpos;
Nlm_RecT true_print_rect, title_rect;
Nlm_FonT title_font;

if(notu == 0) return;
w = Nlm_StartPrinting();
if(w == NULL) return;
Nlm_WatchCursor();
doing_print = TRUE;
Nlm_PrintingRect(&true_print_rect);
title_rect = true_print_rect;
title_font = Nlm_ParseFont(TITLEFONT);
Nlm_SelectFont(title_font);
calc_text_size("Mq", &char_height, &ascent);
title_rect.bottom = title_rect.top + ascent - 1;
true_print_rect.top += char_height;

print_rect.left = true_print_rect.left; print_rect.right = true_print_rect.right;
print_rect.top = true_print_rect.top; print_rect.bottom = true_print_rect.bottom;
h = print_rect.bottom - print_rect.top + 1;
superpos = h / 30;
print_rect.bottom = print_rect.top + (page_count - 1) * (h - superpos) + h - 1;

	totnoms = tottraits = totpoints = -1;
	end_br_length = br_length_txt;
	currx = 0.;
	maxx = 0.;
	nexty = -deltay;
	if(subtree_center == NULL)
		mem_plot(NULL, racine, currx, &curry);
	else
		mem_plot(subtree_ascend, subtree_center, currx, &curry);

for(page = 0; page < page_count; page++) {
	Nlm_StartPage();
	Nlm_ClipPrintingRect( &title_rect);
	print_title(title_rect.left, title_rect.bottom, fname, 
		title_font, page + 1, page_count);
	Nlm_ClipPrintingRect( &true_print_rect);
	Nlm_SelectFont(current_font);
	Nlm_Gray();
	Nlm_FrameRect(&true_print_rect);
	Nlm_Black();
	physx = print_rect.right - print_rect.left;
	physy = print_rect.bottom - print_rect.top;
	do_plot();
	Nlm_EndPage();
	print_rect.top -= (h - superpos); print_rect.bottom -= (h - superpos);
	}
Nlm_EndPrinting(w);
need_runtree = TRUE;
doing_print = FALSE;
ArrowCursor();
}

#endif


#ifdef WIN_MSWIN
void copy_plot(IteM unused)
{
double currx, curry;
Nlm_WindoW w;
Nlm_RecT r;

if(notu == 0) return;
/* dans la picture, les calculs de taille de caractere ne marchent pas!
il faut les faire dans le DC de la fenetre
*/
WatchCursor();
doing_copy = TRUE;
Nlm_GetPosition( fd_nj_plot->tree_plot , &r);
w = Nlm_StartPicture(&r);
totnoms = tottraits = totpoints = -1;
end_br_length = br_length_txt;
currx = 0.;
maxx = 0.;
nexty = -deltay;
Nlm_Black();
Nlm_SelectFont(current_font);
if(subtree_center == NULL)
	mem_plot(NULL, racine, currx, &curry);
else
	mem_plot(subtree_ascend, subtree_center, currx, &curry);
do_plot();
Nlm_EndPicture(w);
need_runtree = TRUE;
doing_copy = FALSE;
ArrowCursor();
}
#endif


char *preptree(char *fname)
{
int i, c, steparbre, maxarbre, maxlname, v;
FILE *njfile;
char *arbre, *der_arbre, *finarbre, *tmp, *p;
static char message[200];

if( (njfile=fopen(fname,"r")) == NULL ){
	sprintf(message, "Tree file %s not found.", fname);
	return (message);
	}
/* recherche du debut de la description de l'arbre */
do c=fgetc(njfile);
while ( isspace(c) );
/* for fastDNAml format, skip initial [comment] */
if(c == '[') {
	do c=fgetc(njfile); while (c != ']');
	do c=fgetc(njfile); while ( isspace(c) );
	}
if ( c != '(') {
	fclose(njfile);
	goto erreur;
	}
/* lecture de l'arbre par paquets de steparbre caracteres*/
	steparbre=1000;
	maxarbre=steparbre;
	arbre=check_alloc(maxarbre,1);
	der_arbre = arbre+maxarbre;
	*arbre=c;
	notu=2; i=3; v = 0;
	finarbre=arbre;
	while( (c=fgetc(njfile)) != EOF && c != ';') {
		if( c=='\n' || c=='\r') continue;
		if(++finarbre >= der_arbre-2) {
			maxarbre += steparbre;
			tmp=check_alloc(maxarbre,1);
			memcpy(tmp,arbre,finarbre-arbre);
			finarbre=tmp+(finarbre-arbre);
			free(arbre);
			arbre=tmp;
			der_arbre=arbre+maxarbre;
			}
		*(finarbre)=c;
		if(c == ')') notu++;
		if(c == '(') i++;
		if(c == ',') v++;
		}
	*(finarbre+1)='\0';
fclose(njfile);
if(i!=notu)goto erreur;

arbre = (char *)realloc(arbre, strlen(arbre) + 4 * v + 5 ); /* worst case add 4 chars for each , */
make_binary_or_unrooted(arbre);
long_arbre_parenth = strlen(arbre);
	notu = 2; p = arbre;
	while( *p != 0 ) {
		if( *p == ')')notu++;
		p++;
		}

notu--; /* after this notu = number of OTUs - 1 for unrooted trees */
totbranches= -1;
/* allocate all memory */
tabtax = (struct noeud **)check_alloc(2*notu +1,sizeof(struct noeud *));
if(notu > 2) branches = (branche *)check_alloc(notu-2,sizeof(branche));
tabnames = (char **)check_alloc(2*notu+1, sizeof(char *));
for(i=0; i<2*notu+1; i++) *(tabtax+i)=
			(struct noeud *)check_alloc(1,s_noeud);
noms = (struct nom *)check_alloc(5*notu+1,sizeof(struct nom));
points = (struct mon_point *)check_alloc(4*notu+1,sizeof(struct mon_point));
traits = (struct trait *)check_alloc(3*notu,sizeof(struct trait));
labels = (char **)check_alloc(notu+1, sizeof(char *));
widnames = (int *)check_alloc(notu+1, sizeof(int));
profs = (double *)check_alloc(notu+1, sizeof(double));
br_length_txt = (char *)check_alloc(2*notu,10);
loadphylip(arbre);
free(arbre);
maxlname = 0; /* largeur max des noms des feuilles */
for(i = 0; i <= nextotu; i++) {
	c = strlen(labels[i]);
	if(c > maxlname) maxlname = c;
	}
for(i = 0; i < 2*notu+1; i++) {
	tabnames[i] = (char *)check_alloc(maxlname + 2 + 1, 1); /*2=place pour # */
	}
	
if(!rooted) {
	racine = *(tabtax+(++num_noeud));
	if(num_noeud >= 2*notu + 1) bad_format("Error: incorrect tree file");
	if(has_br_length) {
		place_midpoint_root();
		root_num = -1;
		}
	else	{
/* ancienne version: derniere espece est groupe externe
*/
		racine->v3 = NULL;
		root_num = notu; 
		}
	}
else	{
	racine = *(tabtax+num_noeud);
	root_br_l= racine->l1 + racine->l2;
	root_num = num_noeud;
	if(!has_br_length) tree_bal(racine,racine->v1,racine->v2);
/* y a-t-il un bootstrap sur l'une des branches racine ? */
	i = get_br_from_bouts(racine, racine->v1); 
	if(i == -1) i = get_br_from_bouts(racine, racine->v2);
	if(i != -1) {
		branches[i].bouta = racine->v1;
		branches[i].boutb = racine->v2;
		}
	}
if(notu+1<3) return ("Tree should contain at least 3 elements.");
subtree_notu = notu;
#ifndef NO_GUI
scrollnotu(notu);
#endif
return NULL;

erreur:
#ifdef WIN_MAC
return ("File or pasted data does not contain correct tree.");
#else
return ("File does not contain correct tree data.");
#endif
} /* end of preptree */


char *check_alloc(int nbrelt, int sizelt)
{
char *retval;
if( (retval=(char *)calloc(nbrelt,sizelt)) != NULL ) return retval;
err_message("ERROR: Not enough memory.");
exit(1);
}


void loadphylip(char *arbre)
{
char *deba,*debb,*debc, *finarbre;
struct noeud *p1, *p2, *p3, *p;
branche *int_br_g, *int_br_d;

has_br_length = 2;
has_internal = FALSE;
/* ignore all stuff after last closing parenthesis 
(needed for fastDNAml output)
*/
finarbre= nextpar(arbre);
rooted=0;
deba=arbre+1;
debb=deba;
while(*debb != ',') {
	if(*debb == 0) bad_format("Incorrect tree file");
	if(*debb == '(')debb=nextpar(debb);
	debb++;
	}
debb++;
debc=debb;
while(*debc != ',' && debc<finarbre) {
	if(*debc == '(')debc=nextpar(debc);
	debc++;
	}
if(*debc==',') {
/* the tree is unrooted <==> it has 3 subtrees at its bottommost level */
	debc++;
	}
else	{
/* the tree is rooted */
	debc=finarbre+1;
	rooted=1;
/*!!!!!!!! notu computed before was 1 unit too high !!!!!!!!*/
	notu--;
	}

num_noeud=notu;
nextotu= -1;
p1=unrootedset(deba,debb-2,&int_br_g);
p2=unrootedset(debb,debc-2,&int_br_d);
p = *(tabtax+(++num_noeud));
if(num_noeud >= 2*notu + 1) bad_format("Error: incorrect tree file");
if(!has_br_length) {
	p1->l3 = 0.5*p1->l3;
	p2->l3 = 0.5*p2->l3;
	}
p->v1=p1; p1->v3=p; p->l1=p1->l3;
if(int_br_g!=NULL) { int_br_g->bouta=p; int_br_g->boutb=p1; }
p->v2=p2; p2->v3=p; p->l2=p2->l3;
if(int_br_d!=NULL) { int_br_d->bouta=p; int_br_d->boutb=p2; }
if(!rooted) {
	p3=unrootedset(debc,finarbre-1,&int_br_g);
	if(int_br_g!=NULL) { int_br_g->bouta=p; int_br_g->boutb=p3; }
	p->v3=p3; p3->v3=p; p->l3=p3->l3;
	}
else	{
	p->v3=NULL;
/* recherche d'un dernier label interne */
	debc=finarbre+1;
	while(*debc!=0 && *debc!=':' && *debc!='[') debc++;
	if(debc-finarbre>1) {
		int l=debc-finarbre-1;
		has_internal = TRUE;
		totbranches++;
		branches[totbranches].br_label=check_alloc(l+1,1);
		memcpy(branches[totbranches].br_label,finarbre+1,l);
		branches[totbranches].br_label[l]=0;
		branches[totbranches].bouta=p1;
		branches[totbranches].boutb=p2;
		}
	}
}


struct noeud *unrootedset(char *deb, char *fin, branche **p_int_br)
{
struct noeud *p, *pp;
char *virg, *ferme;
branche *int_br;
static int l;
static double brlength;

*p_int_br=NULL;
while(*deb==' ')deb++;
while(*fin==' ')fin--;
if(*deb != '(') { /* une feuille */
	virg = strchr(deb, ':');
	if(virg != NULL && virg < fin) {
		if(has_br_length == 0) goto problem;
		sscanf(virg+1, "%le", &brlength);
		has_br_length=1;
		}
	else	{
		if(has_br_length == 1) goto problem;
		brlength = 1;
		has_br_length=0;
		virg = fin + 1;
		}
	virg--;
	while(*deb==' ' || *deb=='\'')deb++;
	if( virg-1 >= deb && *virg == '\'' ) virg--;
	l = virg-deb+1;
	labels[ ++nextotu] = (char *)check_alloc(l + 1, 1);
	memcpy(labels[nextotu], deb, l);
	labels[nextotu][l] = 0;
	p = *(tabtax + nextotu);
	p->l3 = brlength;
	p->v1 = p->v2 = p->v3 = NULL;
	return p;
	}
/* un noeud */
num_noeud++;
if(num_noeud >= 2*notu + 1) bad_format("Error: incorrect tree file");
p = *(tabtax + num_noeud);
ferme =  nextpar(deb);
virg=deb + 1;
while(*virg != ',' && virg < fin) {
	if(*virg == '(') virg=nextpar(virg);
	virg++;
	}
if(virg>=ferme) bad_format("Error: incorrect tree file");
pp = unrootedset(deb + 1, virg - 1, &int_br);
p->v1 = pp; pp->v3 = p; p->l1 = pp->l3;
if(int_br != NULL) { int_br->bouta = p; int_br->boutb = pp; }
pp = unrootedset(virg + 1, ferme - 1, &int_br);
p->v2 = pp; pp->v3 = p; p->l2 = pp->l3;
if(int_br != NULL) { int_br->bouta = p; int_br->boutb = pp; }
virg = strchr(ferme, ':');
if(virg != NULL && virg < fin) { /* traitement longueur */
	if(has_br_length == 0) goto problem;
	sscanf(virg+1, "%le", &brlength);
	has_br_length=1;
	if(*fin == ']') { /* bootstrap entre [] apres longueurs */
		static char *q;
		q = fin - 1;
		while(q > virg && *q != '[') q--;
		if(*q == '[' && fin - q >= 2) {
			has_internal = TRUE;
			totbranches++;
			l = fin - q - 1;
			branches[totbranches].br_label =
				check_alloc(l+1,1);
			memcpy(branches[totbranches].br_label,q+1,l);
			branches[totbranches].br_label[l]=0;
			*p_int_br= &branches[totbranches];
			}
		}
	}
else	{
	if(has_br_length == 1) goto problem;
	brlength = 1;
	has_br_length=0;
	virg = fin + 1;
	}
/* recherche bootstrap (internal label) */
l=virg-ferme-1;
if(l>0) {
	has_internal = TRUE;
	totbranches++;
	branches[totbranches].br_label=
		check_alloc(l+1,1);
	memcpy(branches[totbranches].br_label,ferme+1,l);
	branches[totbranches].br_label[l]=0;
	*p_int_br= &branches[totbranches];
	}
p->l3 = brlength;
return p;

problem:
err_message("Error: Inconsistent tree file for branch lengths.");
exit(1);
}


char *nextpar(char *pospar)
{
char *pos;
pos=pospar+1;
while(*pos != ')') {
	if(*pos == 0) bad_format("Error: unbalanced parentheses");
	if(*pos == '(') pos=nextpar(pos);
	pos++;
	}
return pos;
}


void bad_format(char *mess)
{
err_message(mess);
exit(1);
}



void make_binary_or_unrooted(char *arbre)
{
char *finarbre, *deba, *debb, *debc;

finarbre= nextpar(arbre);
*(finarbre + 1) = 0;
deba=arbre+1;
debb=deba;
while(*debb != ',') {
	if(*debb == 0) bad_format("Incorrect tree file");
	if(*debb == '(')debb=nextpar(debb);
	debb++;
	}
debb++;
debc=debb;
while(*debc != ',' && debc<finarbre) {
	if(*debc == '(')debc=nextpar(debc);
	debc++;
	}
if(*debc == ',') {
/* the tree is unrooted <==> it has 3 subtrees or more at its bottommost level */
	make_binary(arbre , debb, finarbre - 1, TRUE);
	make_binary(arbre , deba, debb - 2, TRUE);
	}
else make_binary(arbre , deba, finarbre - 1, TRUE);
return;
}


int make_binary(char *arbre, char *debut, char *fin, int go_down)
{
int virg, l, retval;
char *p, *q;

p = debut; virg = 0; retval = 0;
while(p < fin) {
	if(*p == ',') virg++;
	else if(*p == '(')	{
		q = nextpar(p);
		if(go_down) {
			l = make_binary(arbre, p+1, q-1, TRUE);
			fin += l;
			retval += l;
			p = q + l;
			}
		else p = q;
		}
	p++;
	}
if(virg > 1) {
	/* recherche de la 2eme virgule */
	p = debut; l = 0;
	while(TRUE) {
		if(*p == ',') {
			l++;
			if(l == 2) break;
			}
		else if(*p == '(')	{
			p = nextpar(p);
			}
		p++;
		}
	l = strlen(p);
	memmove(p + 4, p, l + 1);
	memmove(debut + 1, debut, p - debut);
	*debut = '(';
	memcpy(p + 1, "):0", 3);
	fin += 4;
	retval += 4;
	if(virg > 2) retval += make_binary(arbre, debut, fin, FALSE);  
	}
return retval;
}


void mydrawstring(double x, double y, char *nom, char option)
{
static double px,py;

ch_echelle(x,y,&px,&py);
#ifndef TTY
if(option == '1') {
/* ecrire une chaine en la montant d'un chouia */
	py += char_height/6.;
	}
else if(option == 'c' || option == 'r')	{
/* ecrire une chaine en la centrant vert. sur cette position*/
	py -= char_height/3.;
	px += char_height/6.;
	}
else if(option == 't')	{
	py -=  (5./6.) * char_height;
	}
else if(option == 'b')	{
	py += char_height/6.;
	}
#endif
dir_moveto(px,py);
#ifndef NO_GUI
if(option == 'r') 
	Nlm_Red();
#endif
plotstring(nom);
#ifndef NO_GUI
if(option == 'r') 
	Nlm_Black();
#endif
}


void moveto(double x,double y)
/* move from current pos to logical coord x,y */
{
static double px,py;
ch_echelle(x,y,&px,&py);
dir_moveto(px,py);
}


void lineto(double x,double y)
/* draw line from current pos to logical coord x,y */
{
static double px,py;
ch_echelle(x,y,&px,&py);
dir_lineto(px,py);
}

/* pour un arbre sans longueur de branche, les calculer de facon a ce
que toutes les feuilles arrivent a la meme profondeur */
double tree_bal(struct noeud *centre, struct noeud *p1, struct noeud *p2)
{
double ld, lg, pg, pd, *abup1, *abup2, *abdown1, *abdown2;
struct noeud *f1, *f2;
get_next_br(centre,p1,&abup1,&abdown1,&f1,&f2);
if(f1!=NULL) lg=tree_bal(p1,f1,f2);
else lg=0.0;
get_next_br(centre,p2,&abup2,&abdown2,&f1,&f2);
if(f1!=NULL) ld=tree_bal(p2,f1,f2);
else ld=0.0;
pg=lg+1; pd=ld+1;
if(pg>pd) pd=pg;
else pg=pd;
*abup1 = *abdown1 = pg-lg;
*abup2 = *abdown2 = pd-ld;
return (pg);
}


void get_next_br(struct noeud *pere, struct noeud *fils, 
double **abdown, double **abup, struct noeud **f1, struct noeud **f2)
/* pour une branche pere->fils donnee, calculer dans *f1, *f2 les deux autres
voisins de fils et dans *abdown et *abup les adresses des longeurs de la
branche pere->fils dans les deux sens */
{
if(pere->v1==fils) *abdown= &(pere->l1);
else if(pere->v2==fils) *abdown=&(pere->l2);
else *abdown=&(pere->l3);
if(fils->v1==pere) {
	*abup=&(fils->l1);
	*f1=fils->v2;
	*f2=fils->v3;
	}
else if(fils->v2==pere) {
	*abup=&(fils->l2);
	*f1=fils->v1;
	*f2=fils->v3;
	}
else 	{
	*abup=&(fils->l3);
	*f1=fils->v1;
	*f2=fils->v2;
	}
}


void place_midpoint_root(void)
/* enraciner l'arbre sans racine en cherchant son centre 
*/
{
struct noeud *aux;
double laux;

current_best_diff = VERY_BIG;
if(outotu != NULL) {
	int i; 
	for(i=0; i <= notu; i++) {
		if(labels[i] != NULL && strcmp(labels[i], outotu) == 0) break;	
		}
	if(i > notu) {
		fprintf(stderr, "Error: outotu %s not found in tree\n", outotu);
		exit(1);
		}
	process_branche(tabtax[i], tabtax[i]->v3, tabtax[i]->l3);
	}
else {
	current_cote1=current_cote2=NULL;
	parcourir_branches(*tabtax, NULL);
	}
rooted = TRUE;
root_br_l = current_br_length;
/* il faut toujours que la racine soit telle que racine->v1->v3=racine */
if (current_cote1->v1 == current_cote2 ) {
/* echanger les voisins v1 et v3 de cote1 */
	aux=current_cote1->v1;
	current_cote1->v1=current_cote1->v3;
	current_cote1->v3=aux;
	laux=current_cote1->l1;
	current_cote1->l1=current_cote1->l3;
	current_cote1->l3=laux;
	}
else if (current_cote1->v2 == current_cote2) {
/* echanger les voisins v2 et v3 de cote1 */
	aux=current_cote1->v2;
	current_cote1->v2=current_cote1->v3;
	current_cote1->v3=aux;
	laux=current_cote1->l2;
	current_cote1->l2=current_cote1->l3;
	current_cote1->l3=laux;
	}
current_cote1->v3 = racine;

if (current_cote2->v1 == current_cote1 )
	current_cote2->v1 = racine;
else if (current_cote2->v2 == current_cote1)
	current_cote2->v2 = racine;
else
	current_cote2->v3 = racine;
racine->v1=current_cote1;
racine->v2=current_cote2;
racine->v3=NULL;
racine->l3=0;
/* avoid very unbalanced division of root branch */
if(current_balance > MAX_FRAC) current_balance = MAX_FRAC;
else if(current_balance < 1 - MAX_FRAC) current_balance = 1 - MAX_FRAC;
racine->l1 = current_br_length * current_balance;
racine->l2 = current_br_length - racine->l1;
}


void parcourir_branches(struct noeud *centre, struct noeud *origine)
/* parcourir recursivement toutes les branches de l'arbre sans racine
a partir de centre et dans la direction opposee a son voisin origine
*/
{
if(centre==NULL) return;
if(centre->v1!=origine) {
	process_branche(centre,centre->v1,centre->l1);
	parcourir_branches(centre->v1,centre);
	}
if(centre->v2!=origine) {
	process_branche(centre,centre->v2,centre->l2);
	parcourir_branches(centre->v2,centre);
	}
if(centre->v3!=origine) {
	process_branche(centre,centre->v3,centre->l3);
	parcourir_branches(centre->v3,centre);
	}
}


/* no longer useful */
void process_branche2(struct noeud *cote1, struct noeud *cote2, double length)
/* calculer la prof moyenne des 2 cotes de la branche cote1<-->cote2
de longueur length
et memoriser la meilleure branche dans les variables globales
*/
{
double b1, b2, x, dist_root_side1, dist_root_side2, diff_betw_sides;

if(cote1==NULL || cote2==NULL) return;
if(length <= 0) return;
b1=get_length_down(cote2,cote1);
b2=get_length_down(cote1,cote2);
if( fabs(length) > 1.e-8 )
	x=(b2-b1+length)/(2*length);
else
	x=0;
if(x<0) x=0;
if(x>1) x=1;
dist_root_side1=length*x+b1;
dist_root_side2=length*(1-x)+b2;
diff_betw_sides=fabs(dist_root_side1-dist_root_side2);
if(diff_betw_sides < current_best_diff ) {
	current_best_diff=diff_betw_sides;
	current_cote1=cote1;
	current_cote2=cote2;
	current_br_length=length;
	current_balance=x;
	}
}


void process_branche(struct noeud *cote1, struct noeud *cote2, double length)
/* calculer la variance des distances racine - feuilles si racine est sur branche cote1 - cote2
et memoriser la meilleure branche dans les variables globales
*/
{
double x ; /* partage de branche par facteur x */
moments m1, m2;
double A, B, C; /* variance(x) = A x x + B x + C */
double mini_val; /* meilleure variance pour toutes valeurs de x entre 0 et 1 */

if(cote1 == NULL || cote2 == NULL) return;
if( length < 0 ) length = 0;

m1 = stat_from_node(cote2, cote1);
m2 = stat_from_node(cote1, cote2);
A = 4 * m1.N * ( m2.N * length ) * length;
B = 4 * length * ( m2.N * m1.somme - m1.N * m2.somme - length * m1.N * m2.N);
C = (m1.N + m2.N) * (m1.carres + m2.carres) + m1.N * length * m2.N * length +
	2 * m1.N * length * m2.somme - 2 * m2.N * length * m1.somme -
	(m1. somme + m2.somme) * (m1. somme + m2.somme);
if(A < 1e-20) {
	x = 0.5;
	mini_val = VERY_BIG * 0.99 ;
	}
else {
	x = - B / (2 * A);
	mini_val = C - (B * B) / 4 / A;
	}
if( x < 0 ) {
	x = 0; mini_val = C;
	}
else if( x > 1) {
	x = 1; mini_val = A + B + C;
	}

if(mini_val < current_best_diff ) {
	current_best_diff = mini_val;
	current_cote1 = cote1;
	current_cote2 = cote2;
	current_br_length = length;
	current_balance = x;
	}
}


/* no longer useful */
double get_length_down(struct noeud *pere, struct noeud *racine)
/* compute the average length of the tree down a node */
{
if(racine == NULL) return 0.0;
else	{
	struct noeud *gauche, *droite;
	double bg,bd,lg,ld;
	if( racine->v1 == pere ) {
		gauche =racine->v2; droite = racine->v3;
		bg = (racine->l2); bd = (racine->l3);
		}
	else if( racine->v2 == pere ) {
		gauche =racine->v1; droite = racine->v3;
		bg = (racine->l1); bd = (racine->l3);
		}
	else	{
		gauche =racine->v1; droite = racine->v2;
		bg = (racine->l1); bd = (racine->l2);
		}
 	/* conserver cette ecriture sinon plante sur PC */
 	lg = ld = 0;
 	if(gauche != NULL) lg = get_length_down(racine,gauche);
 	lg += bg;
 	if(droite != NULL) ld = get_length_down(racine,droite);
 	ld += bd;
	return ((lg+ld)/2);
	}
}


moments stat_from_node(struct noeud *pere, struct noeud *racine)
/* compute the moments down a node to all descending tips */
{
struct noeud *gauche, *droite;
moments m;
 static  moments mtmp;
double bg, bd;

	if( racine->v1 == pere ) {
		gauche =racine->v2; droite = racine->v3;
		bg = (racine->l2); bd = (racine->l3);
		}
	else if( racine->v2 == pere ) {
		gauche =racine->v1; droite = racine->v3;
		bg = (racine->l1); bd = (racine->l3);
		}
	else	{
		gauche =racine->v1; droite = racine->v2;
		bg = (racine->l1); bd = (racine->l2);
		}
 	if(gauche != NULL) {
		mtmp = stat_from_node(racine, gauche);
		m.N = mtmp.N; /* nbre de fils */
		m.somme = mtmp.somme + bg * mtmp.N;
		m.carres = mtmp.carres + 2 * bg * mtmp.somme + mtmp.N * bg * bg;

		mtmp = stat_from_node(racine, droite);
		m.N += mtmp.N; /* nbre de fils */
		m.somme += mtmp.somme + bd * mtmp.N;
		m.carres += mtmp.carres + 2 * bd * mtmp.somme + mtmp.N * bd * bd;
 		}
 	else {
 		m.N = 1;
 		m.somme = 0;
 		m.carres = 0;
 		}

return m;
}


void runtree(void)
{
double currx, curry;
int i;
struct noeud *p1, *p2;
double b1,b2,frac_gauche;
if(!rooted) {
	/* place root at user-chosen place: node # root_num */
	rooted = !rooted;
	p1 = *(tabtax+root_num);
	p2 = p1->v3;
	root_br_l = p1->l3;
	if(has_br_length) {
		current_best_diff = VERY_BIG;
		process_branche(p1, p2, root_br_l);
		}
	p1->v3 = racine;
	if (p2->v1 == p1 )
		p2->v1 = racine;
	else if (p2->v2 == p1)
		p2->v2 = racine;
	else
		p2->v3 = racine;
	racine->v1=p1;
	racine->v2=p2;
	racine->v3=NULL;
	if(has_br_length) {
/* old version: by average length
		/* compute the average length of the 2 half trees : b1,b2 *
		b1 = get_length_down(racine,p1);
		b2 = get_length_down(racine,p2);
		/* divide the central branch in 2 parts to equilibrate tree *
		frac_gauche = (root_br_l+b2-b1)/(2*root_br_l);
*/		
		frac_gauche = current_balance;
		if(frac_gauche>MAX_FRAC) frac_gauche=MAX_FRAC;
		else if(frac_gauche<1-MAX_FRAC) frac_gauche=1-MAX_FRAC;
		b1 = frac_gauche*root_br_l;  b2 = root_br_l - b1;
		racine->l1=b1; racine->l2=b2;
		}
	else	{
		tree_bal(racine,p1,p2);
		}
	}
/* initialize leave and node names */
for(i = 0; i <= 2*notu; i++) tabnames[i][0] = 0;
if (choix == depl_racine)
	for(i=0; i<=2*notu-1; i++) sprintf(tabnames[i],"# ");
else if(choix == permutation)
	for(i=notu+1; i<=2*notu; i++) sprintf(tabnames[i],"#");
else if(choix == subtree)
	for(i=notu+1; i<=2*notu-1; i++) sprintf(tabnames[i],"#");

for(i=0; i<=notu; i++) {
	int j;
	j = strlen(tabnames[i]);
	strcpy(tabnames[i] + j, labels[i]);
	}
#ifndef ADDROOT
totnoms = tottraits = totpoints = -1;
end_br_length = br_length_txt;
currx = 0.;
maxx = 0.;
memset(profs, 0, (notu+1)*sizeof(double));
if(subtree_center != NULL)
	deltay = maxy / subtree_notu;
else {
	deltay = maxy / notu;
	subtree_notu = notu;
	}
nexty = -deltay;
calc_text_size("Mq", &char_height, &ascent);
if(subtree_center == NULL)
	mem_plot(NULL, racine, currx, &curry);
else
	mem_plot(subtree_ascend, subtree_center, currx, &curry);
converted = FALSE;
#endif
} /* end of runtree */


void mem_plot(struct noeud *pere, struct noeud *centre, double currx,
double *curry)
/* prepare tree by memorizing all graphic requests */
{
int num=0;
char *p;
struct noeud *gauche, *droite;
double bg, bd, bpere;

while( *(tabtax+num) != centre) num++;
if(pere != NULL && 
	(centre->v1 == NULL || centre->v2 == NULL || centre->v3 == NULL) ) {
	/* orienter le noeud centre de maniere standard: centre->v3=pere */
	if( centre->v1 == pere ) {
		gauche =centre->v2; droite = centre->v3;
		bg = centre->l2; bd = centre->l3; bpere = centre->l1;
		}
	else if( centre->v2 == pere ) {
		gauche =centre->v1; droite = centre->v3;
		bg = centre->l1; bd = centre->l3; bpere = centre->l2;
		}
	else	{
		gauche =centre->v1; droite = centre->v2;
		bg = centre->l1; bd = centre->l2; bpere = centre->l3;
		}
	centre->v3=pere; centre->v1=gauche; centre->v2=droite;
	centre->l3=bpere; centre->l1=bg; centre->l2=bd;
	/* write taxon name */
	nexty += deltay;
	mem_nom(currx,nexty,tabnames[num],'c');
	profs[num] = currx;
	widnames[num] = calc_text_size(tabnames[num], NULL, NULL);
	if(choix==depl_racine) mem_point(currx,nexty,num+1);
	*curry = nexty;
	}
else 	{
	double yg, yd, xg, xd;

	static int doswap;
	static struct noeud *tmp;
	/* doswap vrai ssi permutation de 2 descendants necessaire ici */
	doswap = (swap != 0 && *(tabtax+swap) == centre);
	if( centre->v1 == pere ) {
		if(doswap) {
			tmp= centre->v2;centre->v2= centre->v3;centre->v3= tmp;
			bg= centre->l2; centre->l2= centre->l3; centre->l3= bg;
			}
		gauche =centre->v2; droite = centre->v3;
		bg = centre->l2; bd = centre->l3; bpere = centre->l1;
		}
	else if( centre->v2 == pere ) {
		if(doswap) {
			tmp= centre->v1;centre->v1= centre->v3;centre->v3= tmp;
			bg= centre->l1; centre->l1= centre->l3; centre->l3= bg;
			}
		gauche =centre->v1; droite = centre->v3;
		bg = centre->l1; bd = centre->l3; bpere = centre->l2;
		}
	else	{
		if(doswap) {
			tmp= centre->v1;centre->v1= centre->v2;centre->v2= tmp;
			bg= centre->l1; centre->l1= centre->l2; centre->l2= bg;
			}
		gauche =centre->v1; droite = centre->v2;
		bg = centre->l1; bd = centre->l2; bpere = centre->l3;
		}
	/* orienter le noeud centre de maniere standard: centre->v3=pere */
	centre->v3=pere; centre->v1=gauche; centre->v2=droite;
	centre->l3=bpere; centre->l1=bg; centre->l2=bd;
	xg=currx+bg; xd=currx+bd;

	mem_plot(centre,gauche,xg,&yg);
	mem_plot(centre,droite,xd,&yd);

	mem_trait(currx,yg,xg,yg);
	mem_trait(currx,yd,xd,yd);
	mem_trait(currx,yg,currx,yd);
	*curry = (yg+yd)/2;

/* write internal labels */
	if(choix==show_tree && show_bootstrap) {
		if( (p=get_br_label(centre, gauche)) != NULL )
			mem_nom(currx,yg,p,'t');
		if( (p=get_br_label(centre, droite)) != NULL )
			mem_nom(currx,yd,p,'b');
		if(pere==NULL && (p=get_br_label(gauche, droite)) != NULL )
			mem_nom(currx,*curry,p,'c');
		}

/* write node number */
	mem_nom(currx,*curry,tabnames[num],'c');
	mem_point(currx,*curry,num+1);
	if(plot_br_l && has_br_length) {
/* write branch length */
		double min_br_l = 0.008; /* minimum branch length displayed */
		if(bg>min_br_l) {
			sprintf(end_br_length,"%.3f",bg);
			mem_nom(currx+bg/10,yg,end_br_length,'1');
			end_br_length += (strlen(end_br_length)+1);
			}
		if(bd>min_br_l) {
			sprintf(end_br_length,"%.3f",bd);
			mem_nom(currx+bd/10,yd,end_br_length,'1');
			end_br_length += (strlen(end_br_length)+1);
			}
		}
	}
}  /* end of mem_plot */

void mem_point(double x, double y, int number)
{
++totpoints;
(points+totpoints)->x = x;
(points+totpoints)->y = y;
(points+totpoints)->number = number;
}

void mem_nom(double x, double y, char *nom, char option)
/* x,y: logical coordinates of beginning of string 
   nom: address of string to be displayed later
   option: 'c' use for position of center of character height
           '1' put bottom of characters at y+1 pixel
	   't' use for position of top of character
	   'b' use for position of bottom of character
*/
{
if(strlen(nom) != 0) {
	totnoms++;
	if(x > maxx) maxx = x;
	(noms+totnoms)->x = x;
	(noms+totnoms)->y = y;
	(noms+totnoms)->nom = nom;
	(noms+totnoms)->disp_option = option;
	}
}


void mem_trait(double xd, double yd, double xf, double yf)
{
tottraits++;
if(xd>maxx) maxx=xd;
if(xf>maxx) maxx=xf;
(traits+tottraits)->xd = xd;
(traits+tottraits)->yd = yd;
(traits+tottraits)->xf = xf;
(traits+tottraits)->yf = yf;
}


char *get_br_label(struct noeud *a, struct noeud *b)
{
int i;
for(i=0; i<=totbranches; i++) {
	if(branches[i].bouta==a && branches[i].boutb==b) 
		return branches[i].br_label;
	if(branches[i].boutb==a && branches[i].bouta==b) 
		return branches[i].br_label;
	}
return NULL;
}


int get_br_from_bouts(struct noeud *a, struct noeud *b)
{
int i;
for(i=0; i<=totbranches; i++) {
	if(branches[i].bouta==a && branches[i].boutb==b) 
		return i;
	if(branches[i].boutb==a && branches[i].bouta==b) 
		return i;
	}
return -1;
}


void free_tree(void)
{
int i;
if(notu == 0) return;
/* de-allocate all memory */
for(i=0; i<2*notu+1; i++)
	free(tabtax[i]);
free(tabtax);
for(i=0; i<notu-2; i++)
	if(branches[i].br_label != NULL) free(branches[i].br_label);
free(branches);
for(i = 0; i < 2 * notu + 1; i++) free(tabnames[i]);
free(tabnames);
free(noms);
free(points);
free(traits);
for(i = 0; i < notu + 1; i++) free(labels[i]);
free(labels);
free(br_length_txt);
}


/* ecriture d'un arbre racine au format phylip */
char *ecrit_arbre_parenth(struct noeud *root)
{
char *arbre, *fin, *p;
int l, maxarbre = 2 * long_arbre_parenth + 1000;

arbre=check_alloc( maxarbre+20,1);
fin=recur_ecrit_arbre(root,arbre,arbre+maxarbre-1);
/* ecriture du dernier label interne */
if( fin != NULL && (p=get_br_label(root->v1,root->v2)) != NULL ) {
	l=strlen(p);
	if(fin+l>=arbre+maxarbre) fin= NULL;
	else	{
		memcpy(fin+1,p,l);
		fin+=l;
		}
	}
if(fin == NULL) {
	free(arbre);
	return NULL;
	}
strcpy(fin+1,";");
return arbre;
}


char *recur_ecrit_arbre(struct noeud *centre, char *arbre, char *finarbre)
{
int num, l;
char *p;

if(centre->v1==NULL && centre->v2==NULL) {
	num=0;
	while( *(tabtax+num) != centre) num++;
	l=strlen(tabnames[num]);
	if(arbre+l>=finarbre) return NULL;	
	memcpy(arbre,tabnames[num],l);
	arbre += l-1;
	}
else	{
	*arbre='(';
	arbre=recur_ecrit_arbre(centre->v1,arbre+1,finarbre);
	if(arbre==NULL) return NULL;
	if(has_br_length) {
		if(arbre+10>=finarbre) return NULL;
		sprintf(++arbre,":%.5f",centre->l1);
		while(*arbre!=0) arbre++;
		}
	else arbre++;
	*arbre=',';
	arbre=recur_ecrit_arbre(centre->v2,arbre+1,finarbre);
	if(arbre==NULL) return NULL;
	if(has_br_length) {
		if(arbre+10>=finarbre) return NULL;
		sprintf(++arbre,":%.5f",centre->l2);
		while(*arbre!=0) arbre++;
		}
	else arbre++;
	*arbre=')';
	/* ecriture des labels internes */
	if( (p=get_br_label(centre,centre->v3))!=NULL ) {
		l=strlen(p);
		if(arbre+l>=finarbre) return NULL;
		memcpy(arbre+1,p,l);
		arbre+=l;
		}
	}
return arbre;
}


void removeroot(void)
{
struct noeud *p1, *p2;
p1=racine->v1;
p2=racine->v2;
if(p1->v1 == racine )
	{p1->v1 = p2; p1->l1 = root_br_l;}
else if (p1->v2 == racine)
	{p1->v2 = p2; p1->l2 = root_br_l;}
else
	{p1->v3 = p2; p1->l3 = root_br_l;}
if(p2->v1 == racine )
	{p2->v1 = p1; p2->l1 = root_br_l;}
else if (p2->v2 == racine)
	{p2->v2 = p1; p2->l2 = root_br_l;}
else
	{p2->v3 = p1; p2->l3 = root_br_l;}
rooted = 0;
}


int myrint(double val)
{
if(val >= 0)
	return (int)(val + 0.5);
else
	return (int)(val - 0.5);
}


int calc_n_desc(struct noeud *pere)
{
if(pere->v1 == NULL) return 1;
else return calc_n_desc(pere->v1) + calc_n_desc(pere->v2);
}


void postscript_plot(void)
{
double currx, curry;
int h, page, superpos, erreur = FALSE, offset;
int paper_h, sheets_h, recouvre, font_size;
const int a4_h = 750, letter_h = 700;
time_t heure;
float screen_w, ps_w;
static int first = TRUE;
const char alphabet[]="ABCDEFGHIJKLMNOPQRSTUVWXYZ";
/* partie dessinable 500(ps_width) x a4_h decalee de 50 x 50 par rapport a coin inferieur gauche du papier
cadre exterieur de 10 sur les 4 cotes, correspond aussi a zone de clip
*/

if(notu == 0) return;
if(paper_choice == A4)
	paper_h = a4_h;
else
	paper_h = letter_h;
if(ps_height != 0) paper_h = ps_height;

#ifndef NO_GUI
Nlm_WatchCursor();
#endif
ps_plot = TRUE;
font_size = list_font_size[current_font_size_rank];

if(first) {
	first = FALSE;
#ifndef NO_GUI
#ifdef WIN_MSWIN
	Nlm_SelectFont( Nlm_ParseFont( "Courier New,12" ) );
#else
	Nlm_SelectFont( Nlm_ParseFont( "Courier,12" ) );
#endif
	screen_w = Nlm_TextWidth(alphabet, 26);/*taille sur ecran*/
	ps_w = 12 * 0.60 * 26; /*taille en postscript pour courier,12*/
	postscript_ratio = ps_w / screen_w ;
	Nlm_SelectFont(current_font);/*retour a police courante pour ecran */
#endif
	}

/* recouvre = paper_h / 30; */
recouvre = 0;
sheets_h = (paper_h - recouvre) * page_count + recouvre;
offset = paper_h - recouvre;
physx = ps_width;
physy = sheets_h;

plotfile=fopen(plotfilename,"w");
fprintf(plotfile,"%%!PS-Adobe-1.0\n"
	"%%%%DocumentFonts: Times-Roman %s\n"
	"%%%%Creator: njplot\n"
	"%%%%Pages: %d\n"
	"%%%%Title: %s\n"
	"%%%%EndComments\n",
	current_ps_font, page_count, fname);
fprintf(plotfile,
	"/setpacking where {true setpacking} if\n"
	"1 setlinecap 1 setlinejoin 1 setlinewidth 0 setgray\n"
	"/basefont /%s findfont %d scalefont def\n",
	current_ps_font, font_size);
fprintf(plotfile,"/titlefont /Times-Roman findfont 12 scalefont def\n");
fprintf(plotfile, "/setclip {40 40 moveto %d 40 lineto %d %d lineto 40 %d "
	"lineto closepath clip newpath} def\n", 
	ps_width + 60, ps_width + 60, paper_h+60, paper_h+60);
time(&heure);
fprintf(plotfile, "/title {titlefont setfont\n"
	"40 %d moveto (%s   %s) show (  Page ) show show ( of %d) show\n"
	"} def\n", 
	paper_h+65, extract_filename(fname), ctime(&heure), page_count);
fprintf(plotfile,"%%%%EndProlog\n");

totnoms = tottraits = totpoints = -1;
end_br_length = br_length_txt;
currx = 0.;
maxx = 0.;
nexty = -deltay;
if(subtree_center == NULL)
	mem_plot(NULL, racine, currx, &curry);
else
	mem_plot(subtree_ascend, subtree_center, currx, &curry);
for(page = 0; page < page_count ; page++) {
	fprintf(plotfile,"%%%%Page: ? %d\n", page+1);
	if( ! no_title) fprintf(plotfile,"(%d) title ", page+1);
	fprintf(plotfile,"setclip\n");
	fprintf(plotfile,"0 %d translate\n", - (page_count - page - 1) * offset);
	fprintf(plotfile,"basefont setfont\n");
	fprintf(plotfile,"50 50 translate\n");
	fprintf(plotfile,"0.7 setgray -10 -10 moveto %d -10 lineto %d %d lineto "
		 "-10 %d lineto closepath stroke 0 setgray \n", 
		 ps_width + 10, ps_width + 10, sheets_h+10, sheets_h+10);
	do_plot();
	fprintf(plotfile,"showpage\n");
	}
fprintf(plotfile, "%%%%Trailer\n");
if(ferror(plotfile)) erreur = TRUE;
if(fclose(plotfile) != 0) erreur = TRUE;

need_runtree = TRUE;
ps_plot = FALSE;
#ifndef NO_GUI
ArrowCursor();
if(!ps_plot_only ) {		
	char mess[250];
	if(!erreur  ) {		
		sprintf(mess, 
			"Tree plot is now in file %s in Postscript format", 
			extract_filename(plotfilename) );
		}
	else {
		sprintf(mess, "Error while writing to file %s", 
			extract_filename(plotfilename) );
		}
	err_message(mess);
	}
#endif
}



void tty_plot(void)
{
double currx, curry;
int h, erreur = FALSE;
char *p;

if(notu == 0) return;
ps_plot = TRUE;

/* physx was set in process_args */
physy = 2*(notu + 1) + 4;
tty_page = (char **)malloc(physy*sizeof(char *));
for(h=0; h < physy; h++) {
	tty_page[h] = (char *)malloc(physx + 1);
	memset(tty_page[h], ' ', physx);
	tty_page[h][(int)physx] = 0;
	}

totnoms = tottraits = totpoints = -1;
end_br_length = br_length_txt;
currx = 0.;
maxx = 0.;
nexty = -deltay;
if(subtree_center == NULL)
	mem_plot(NULL, racine, currx, &curry);
else
	mem_plot(subtree_ascend, subtree_center, currx, &curry);
do_plot();
for(h=0; h < physy; h++) {
	p = tty_page[h] + strlen(tty_page[h]) - 1;
	while(p > tty_page[h] && *p == ' ') *(p--) = 0;
	}

plotfile=fopen(plotfilename, "w");
for(h=1; h <= physy; h++) {
	fprintf(plotfile, "%s\n", tty_page[(int)physy - h]);
	}
if(ferror(plotfile)) erreur = TRUE;
if(fclose(plotfile) != 0) erreur = TRUE;

need_runtree = TRUE;
ps_plot = FALSE;
}


void majuscules(char *p)
{
while(*p != 0) {*p = toupper(*p); p++;}
}


#ifdef WIN_MAC
#define DIR_DELIM ':'
#elif defined(WIN_MSWIN)
#define DIR_DELIM '\\'
#else
#define DIR_DELIM '/'
#endif

char *extract_filename(char *name)
{
char *p;
while( ( p = strchr(name, DIR_DELIM) ) != NULL) name = p + 1;
return name;
}


#ifdef WIN_MOTIF
#include <ncbiport.h>


static void Nlm_GetFontData (Nlm_FonT f, Nlm_FontData * fdata)

{
  Nlm_FntPtr  fp;

  if (f != NULL && fdata != NULL) {
    fp = (Nlm_FntPtr) Nlm_HandLock (f);
    *fdata = *fp;
    Nlm_HandUnlock (f);
  }
}


void set_systemfont(char *fname)
{
		XFontStruct   *font;
		XmFontListEntry entry;
		Nlm_FontData fontdata;
		extern XmFontList Nlm_XfontList;
		
		Nlm_systemFont = Nlm_ParseFont(fname);
        Nlm_GetFontData(Nlm_systemFont, &fontdata);
        font= (XFontStruct *)fontdata.handle;
		entry = XmFontListEntryCreate(XmFONTLIST_DEFAULT_TAG, 
					XmFONT_IS_FONT, (XtPointer)font);
        Nlm_XfontList = XmFontListAppendEntry(NULL, entry);
                
        Nlm_SelectFont(Nlm_systemFont);
        Nlm_stdLineHeight=Nlm_LineHeight();
        Nlm_stdFontHeight=Nlm_FontHeight();
        Nlm_stdAscent=Nlm_Ascent();
        Nlm_stdDescent=Nlm_Descent();
        Nlm_stdLeading=Nlm_Leading();
        Nlm_stdCharWidth = Nlm_MaxCharWidth ();
}
#endif
